"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import subprocess

from pathlib import Path

from gillespy2.core import log
from gillespy2.core import gillespyError

class Make():
    def __init__(self, makefile: str, output_dir: str, obj_dir: str = None):
        self.makefile = Path(makefile).resolve()
        self.self_dir = Path(__file__).parent

        self.cbase_dir = self.self_dir.joinpath("../c_base").resolve()
        self.output_dir = Path(output_dir).resolve()

        # obj_dir is, presumably, only supplied if caching is enabled.
        # If not supplied, it should be assumed ethereal and cleaned up
        #  with the rest of the tmp directory.
        if obj_dir is None:
            self.obj_dir = self.output_dir.joinpath("obj").resolve()
        else:
            self.obj_dir = Path(obj_dir).resolve()

        for path in [self.output_dir, self.obj_dir]:
            if path.is_file():
                raise gillespyError.DirectoryError(
                    f"Error while attempting to open directory:\n"
                    f"- {path} is actually a file."
                )

            if not path.is_dir():
                path.mkdir()

        self.output_file = "Simulation.out"
        if os.name == "nt":
            self.output_file = "Simulation.exe"

        self.output_file = Path(self.output_dir, self.output_file)

    def prebuild(self):
        self.__execute("build")

    def build_simulation(self, simulation_name: str, **kwargs):
        self.__execute(simulation_name, **kwargs)

    def __execute(self, target: str, **kwargs):        
        # Default make arguments.
        args_dict = {
            "cbase_dir": str(self.cbase_dir.resolve()),
            "obj_dir": str(self.obj_dir.resolve()),
            "output_dir": str(self.output_dir.resolve()),
            "output_file": str(self.output_dir.joinpath(self.output_file).resolve())
        }

        # Overwrite keys supplied in **kwargs.
        for key, value in kwargs.items():
            args_dict[key] = value

        # Create the emake arguments. Note, arguments are UPPERCASE.
        make_args = [(f"{key.upper()}={value}") for key, value in args_dict.items()]

        # Create the make command.
        make_cmd = ["make", "-C", str(self.cbase_dir), "-f", str(self.makefile), target] + make_args

        try:
            result = subprocess.run(make_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        except KeyboardInterrupt:
            log.warning(f"Makefile was interrupted during execution of target: '{target}', unexpected behavior may occur.")

        if result.returncode == 0:
            return

        raise gillespyError.BuildError(f"Error encountered during execution of Makefile target: '{target}'.\n"
            f"Return code: {result.returncode}"
            f"- stdout: {result.stdout.decode('utf-8', errors='ignore')}\n"
            f"- stderr: {result.stderr.decode('utf-8', errors='ignore')}\n")


"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import ast  # for dependency graphing
import numpy as np
from gillespy2.core import log, Species

"""
NUMPY SOLVER UTILITIES BELOW
"""


def numpy_initialization(model):
    species_mappings = model.sanitized_species_names()
    species = list(species_mappings.keys())
    parameter_mappings = model.sanitized_parameter_names()
    number_species = len(species)
    return species_mappings, species, parameter_mappings, number_species


def numpy_trajectory_base_initialization(model, number_of_trajectories, timeline, species, resume=None):
    trajectory_base = np.zeros((number_of_trajectories, timeline.size, len(species) + 1))

    # copy time values to all trajectory row starts
    trajectory_base[:, :, 0] = timeline
    tmpSpecies = {}
    # copy initial populations to base
    if resume is not None:
        # Set initial values of species to where last left off
        for i in species:
            tmpSpecies[i] = resume[i][-1]
        for i, s in enumerate(species):
            trajectory_base[:, 0, i + 1] = tmpSpecies[s]
    else:
        for i, s in enumerate(species):
            trajectory_base[:, 0, i + 1] = model.listOfSpecies[s].initial_value

    return trajectory_base, tmpSpecies


def numpy_resume(timeStopped, simulation_data, resume=None):
    """
    Helper function for when resuming a simulation in a numpy based solver.

    :param timeStopped: The time in which the simulation was stopped.
    :param simulation_data: The current models simulation data, after being parsed in the numpy solver of choice.

    :param resume: The previous simulations data, that is being resumed
    :type resume: gillespy2.core.Results

    :returns: Combined simulation data, the old resume data and the current simulation data.
    """
    if timeStopped != 0:
        if timeStopped != simulation_data[0]['time'][-1]:
            tester = np.where(simulation_data[0]['time'] > timeStopped)[0].size
            index = np.where(simulation_data[0]['time'] == timeStopped)[0][0]
        if tester > 0:
            for i in simulation_data[0]:
                simulation_data[0][i] = simulation_data[0][i][:index]
    if resume is not None:
        # If resuming, combine old pause with new data, and delete any excess null data
        for i in simulation_data[0]:
            oldData = resume[i][:-1]
            newData = simulation_data[0][i]
            simulation_data[0][i] = np.concatenate((oldData, newData), axis=None)

    return simulation_data

"""
VARIABLE SOLVER METHODS
"""


def update_species_init_values(listOfSpecies, species, variables, resume = None):
    # Update Species Initial Values
    populations = ''
    for i in range(len(species) - 1):
        if species[i] in variables:
            populations += '{} '.format(int(variables[species[i]]))
        else:
            if resume is not None:
                populations += '{} '.format(int(resume[species[i]][-1]))
            else:
                populations += '{} '.format(int(listOfSpecies[species[i]].initial_value))
    if species[-1] in variables:
        populations += '{}'.format(int(variables[species[-1]]))
    else:
        if resume is not None:
            populations += '{} '.format(int(resume[species[-1]][-1]))
        else:
            populations += '{}'.format(int(listOfSpecies[species[-1]].initial_value))
    return populations

def change_param_values(listOfParameters, parameters, volume, variables):
    # Update Parameter Values
    parameter_values = ''
    for i in range(len(parameters) - 1):
        if parameters[i] in variables:
            parameter_values += '{} '.format(variables[parameters[i]])
        else:
            if parameters[i] == 'vol':
                parameter_values += '{} '.format(volume)
            else:
                parameter_values += '{} '.format(listOfParameters[parameters[i]].expression)
    if parameters[-1] in variables:
        parameter_values += '{}'.format(variables[parameters[-1]])
    else:
        if parameters[-1] == 'vol':
            parameter_values += '{}'.format(volume)
        else:
            parameter_values += '{}'.format(listOfParameters[parameters[-1]].expression)
    return parameter_values

"""
Below are two functions used for creating dependency graphs in the C solvers, and Numpy Solvers.
"""

def species_parse(model, custom_prop_fun):
    """
    This function uses Pythons AST module to parse custom propensity function, looking for Species in a model

    :param model: Model to be checked for species
    :param custom_prop_fun: The custom propensity function to be parsed

    :returns: List of species objects that are found in a custom propensity function
    """
    parsed_species = []

    class SpeciesParser(ast.NodeTransformer):
        def visit_Name(self, node):
            if isinstance(model.get_element(node.id), Species):
                parsed_species.append(model.get_element(node.id))

    expr = custom_prop_fun
    expr = ast.parse(expr, mode='eval')
    expr = SpeciesParser().visit(expr)
    return parsed_species


def dependency_grapher(model, reactions):
    """
    This function returns a dependency graph for a models reactions in the form of a
    dictionary containing {species name: {'dependencies'}:[list of reaction names]}.

    :param model: Model to used to create a reaction dependency graph
    :param reactions: list[model.listOfReactions]

    :returns: Dependency graph dictionary
    """
    dependent_rxns = {}
    for i in reactions:
        cust_spec = []
        if model.listOfReactions[i].type == 'customized':
            cust_spec = (species_parse(model, model.listOfReactions[i].propensity_function))

        for j in reactions:

            if i not in dependent_rxns:
                dependent_rxns[i] = {'dependencies': []}
            if j not in dependent_rxns:
                dependent_rxns[j] = {'dependencies': []}
            if i == j:
                continue

            reactantsI = list(model.listOfReactions[i].reactants.keys())
            reactantsJ = list(model.listOfReactions[j].reactants.keys())

            if j not in dependent_rxns[i]['dependencies']:
                if any(elem in reactantsI for elem in reactantsJ):
                    if i not in dependent_rxns[j]['dependencies']:
                        dependent_rxns[j]['dependencies'].append(i)
                    dependent_rxns[i]['dependencies'].append(j)

            if i not in dependent_rxns[j]['dependencies']:
                if any(elem in list(model.listOfReactions[i].products.keys()) for elem in
                       list(model.listOfReactions[j].reactants.keys())):
                    dependent_rxns[j]['dependencies'].append(i)

            if cust_spec:
                if any(elem in cust_spec for elem in list(model.listOfReactions[j].reactants)) or any \
                            (elem in cust_spec for elem in list(model.listOfReactions[j].products)):
                    dependent_rxns[i]['dependencies'].append(j)

    return dependent_rxns




"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import unittest
from gillespy2.core.gillespyError import SimulationError
from example_models import Example, ExampleNoTspan
from gillespy2 import NumPySSASolver


class TestNumPySSASolver(unittest.TestCase):
    model = Example()

    def test_run_example__with_increment_only(self):
        model = ExampleNoTspan()
        results = NumPySSASolver.run(model=model, increment=0.2)

    def test_run_example__with_tspan_only(self):
        model = Example()
        results = NumPySSASolver.run(model=model)

    def test_run_example__with_tspan_and_increment(self):
        with self.assertRaises(SimulationError):
            model = Example()
            results = NumPySSASolver.run(model=model, increment=0.2)
    

if __name__ == '__main__':
    unittest.main()



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import random
import tempfile
import os
import uuid
import subprocess
import shutil
from gillespy2.core import GillesPySolver, Model, log
from gillespy2.core.gillespyError import SimulationError, InvalidModelError


class StochKitBaseSolver(GillesPySolver):
    name = "StochKitBaseSolver"
    rc = 0
    """
    Abstract class for a StochKit solver. This is generally called from within a
    GillesPy Model through the Model.run function. Returns simulation
    trajectories.


    :param model: The model on which the solver will operate
    :type model: GillesPy2.model
    
    :param t: The end time of the solver
    :type t: float
    
    :param number_of_trajectories: The number of times to sample the chemical master equation. Each
    trajectory will be returned at the end of the simulation.
    :type number_of_trajectories: int
    
    :param increment: The time step of the solution
    :type increment: float
    
    :param seed: The random seed for the simulation. Defaults to None.
    :type seed: int
    
    :param stochkit_home: Path to stochkit. This is set automatically upon installation, but
    may be overwritten if desired.
    :type stochkit_home: str
    
    :param algorithm: The solver by which to simulate the model. 'ssa' or 'tau_leaping'
    are the available options. If 'ssa' is chosen, StochKit will choose
    from the available ssa options.
    :type algorithm: str
    
    :param job_id: If given, this will be the name of the solver run. Usually not set.
    :type job_id: str
    
    :param extra_args:  Any extra arguments for the stochkit solver. See StochKit2
    documentation for details.
    :type extra_args: str
    
    :param debug: Set to True to provide additional debug information about the simulation.
    :type debug: bool
    
    :param show_labels: Use names of species as index of result object rather than position numbers.
    :type show_labels: str
    """
    @classmethod
    def run(cls, model, t=20, number_of_trajectories=1, increment=0.05, seed=None,
            stochkit_home=None, algorithm=None, job_id=None, extra_args='',
            debug=False, profile=False, show_labels=False, **kwargs):
        """
        Call out and run the solver. Collect the results.
        """
        if len(kwargs) > 0:
            for key in kwargs:
                log.warning('Unsupported keyword argument to solver: {0}'.format(key))

        if algorithm is None:
            raise SimulationError("No algorithm selected")

        # We write all StochKit input and output files to a temporary folder
        prefix_base_dir = tempfile.mkdtemp()
        prefix_out_dir = os.path.join(prefix_base_dir, 'output')
        os.mkdir(prefix_out_dir)

        if job_id is None:
            job_id = str(uuid.uuid4())

        if isinstance(model, Model):
            # Write a temporary StochKit2 input file.
            outfile = os.path.join(prefix_base_dir, "temp_input_{0}.xml".format(job_id))
            with open(outfile, 'w') as model_file_handle:
                model_file_handle.write(model.serialize())
        elif isinstance(model, str):
            outfile = model
        else:
            raise InvalidModelError('Model must be either a GillesPy Model instance or an xml file name.')

        executable = cls.locate_executable(stochkit_home=stochkit_home, algorithm=algorithm)

        if executable is None:
            raise SimulationError("stochkit executable '{0}' not found. \
                Make sure it is your path, or set STOCHKIT_HOME environment \
                variable'".format(algorithm))

        # Assemble argument list for StochKit
        out_dir = os.path.join(prefix_out_dir, job_id)

        if increment is None:
            increment = t / 20.0
        num_output_points = round(t / increment)

        # Assemble the argument list
        args = '--model {0} --out-dir {1} -t {2} -i {3}'.format(outfile, out_dir, t, int(num_output_points))

        directories = os.listdir(prefix_out_dir)
        if os.path.isdir(out_dir):
            if debug:
                print('Ensemble {0} already existed, using --force.'.format(job_id))
            args += ' --force'

        # If we are using local mode, shell out and run StochKit
        # (SSA or Tau-leaping or ODE)
        cmd = ' '.join([executable, args, extra_args])
        if debug:
            print("cmd: {0}".format(cmd))

        # Execute
        try:
            handle = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            return_code = handle.wait()
        except OSError as e:
            raise SimulationError("Solver execution failed: {0}\n{1}".format(cmd, e))

        try:
            stderr = handle.stderr.read()
        except Exception as e:
            stderr = 'Error reading stderr: {0}'.format(e)
        try:
            stdout = handle.stdout.read()
        except Exception as e:
            stdout = 'Error reading stdout: {0}'.format(e)

        if return_code != 0:
            raise SimulationError("Solver execution failed: '{0}' output: {1}{2}".format(cmd, stdout, stderr))

        try:
            # Get data using solver specific function
            trajectories = cls.get_trajectories(out_dir, debug=debug, show_labels=show_labels)
            if len(trajectories) == 0:
                raise SimulationError("Solver execution failed: '{0}' output: {1}{2}".format(cmd, stdout, stderr))
            if show_labels:
                labels, trajectories = trajectories
                trajectories = cls.label_trajectories(trajectories, labels)
            return trajectories, cls.rc
        except Exception as e:
            compile_log_file = os.path.join(prefix_base_dir, 'temp_input_{0}_generated_code'.format(job_id),
                                            'compile-log.txt')
            log_file = os.path.join(prefix_out_dir, job_id, 'log.txt')
            for file_name in [compile_log_file, log_file]:
                if os.path.isfile(file_name):
                    with open(file_name) as f:
                        error = f.read()
                    raise SimulationError("Error running simulation: {0}\n{1}\n".format(file_name, error))

            raise SimulationError("Error using solver.get_trajectories('{0}'): {1}".format(out_dir, e))
        finally:
            # Clean up
            if debug:
                print("prefix_base_dir={0}".format(prefix_base_dir))
                print("STDOUT: {0}".format(stdout))
                print("STDERR: {0}".format(stderr))
            else:
                shutil.rmtree(prefix_base_dir)

    @staticmethod
    def locate_executable(stochkit_home=None, algorithm=None):
        # Algorithm, SSA or Tau-leaping?
        executable = None
        if stochkit_home is not None:
            if os.path.isfile(os.path.join(stochkit_home, algorithm)):
                executable = os.path.join(stochkit_home, algorithm)
            else:
                raise SimulationError("stochkit executable '{0}' not found \
                stochkit_home={1}".format(algorithm, stochkit_home))
        elif os.environ.get('STOCHKIT_HOME') is not None:
            if os.path.isfile(os.path.join(os.environ.get('STOCHKIT_HOME'),
                                           algorithm)):
                executable = os.path.join(os.environ.get('STOCHKIT_HOME'),
                                          algorithm)
        if executable is None:
            # try to find the executable in the path
            if os.environ.get('PATH') is not None:
                for directory in os.environ.get('PATH').split(os.pathsep):
                    if os.path.isfile(os.path.join(directory, algorithm)):
                        executable = os.path.join(directory, algorithm)
                        break
        return executable

    @staticmethod
    def process_seed(seed):
        if seed is None:
            seed = random.randint(0, 2147483647)
        # StochKit breaks for long ints
        if seed.bit_length() >= 32:
            seed = seed & ((1 << 32) - 1)
            if seed > (1 << 31) - 1:
                seed -= 1 << 32
        return seed

    @staticmethod
    def label_trajectories(trajectories, labels):
        results = []
        for r in trajectories:
            ret = {}
            for n, l in enumerate(labels):
                ret[l] = r[:, n]
            results.append(ret)
        return results


class StochKitSolver(StochKitBaseSolver):
    name = 'StochKitSolver'
    """
    Abstract class for StochKit solver derived from the GillesPySolver class.
    This is generally used to set up the solver.


    :param model: The model on which the solver will operate
    :type model: GillesPy2.model

    :param t: The end time of the solver
    :type t: float

    :param number_of_trajectories: The number of times to sample the chemical master equation. Each
    trajectory will be returned at the end of the simulation.
    :type number_of_trajectories: int

    :param increment: The time step of the solution
    :type increment: float

    :param seed: The random seed for the simulation. Defaults to None.
    :type seed: int

    :param stochkit_home: Path to stochkit. This is set automatically upon installation, but
    may be overwritten if desired.
    :type stochkit_home: str

    :param algorithm: The solver by which to simulate the model. 'ssa' or 'tau_leaping'
    are the available options. If 'ssa' is chosen, StochKit will choose
    from the available ssa options.
    :type algorithm: str

    :param job_id: If given, this will be the name of the solver run. Usually not set.
    :type job_id: str

    :param extra_args:  Any extra arguments for the stochkit solver. See StochKit2
    documentation for details.
    :type extra_args: str

    :param debug: Set to True to provide additional debug information about the simulation.
    :type debug: bool

    :param show_labels: Use names of species as index of result object rather than position numbers.
    :type show_labels: str
    """
    @classmethod
    def run(cls, model, t=20, number_of_trajectories=1, increment=0.05, seed=None,
            stochkit_home=None, algorithm='ssa', job_id=None, method=None,
            debug=False, show_labels=False, profile=False, processes=1, **kwargs):

        # all this is specific to StochKit
        if model.units == "concentration":
            raise SimulationError("StochKit can only simulate population \
                                  models, please convert to population-based model for \
                                  stochastic simulation. Use solver = StochKitODESolver \
                                  instead to simulate a concentration model deterministically.")

        seed = super().process_seed(seed)

        # We keep all the trajectories by default.
        args = ' -p {0} --keep-trajectories --label --seed {1} --realizations {2}'.format(processes, seed, number_of_trajectories)

        if method is not None:  # This only works for StochKit 2.1
            args += ' --method ' + str(method)

        return super().run(model=model, t=t, number_of_trajectories=number_of_trajectories, increment=increment, seed=seed, stochkit_home=stochkit_home,
                           algorithm=algorithm, job_id=job_id, debug=debug, show_labels=show_labels, extra_args=args)

    @classmethod
    def get_trajectories(cls, out_dir, debug=False, show_labels=False):
        if debug:
            print("StochKitSolver.get_trajectories(out_dir={0}".format(out_dir))
        # Collect all the output data
        trajectories = []
        trajectory_directory = os.path.join(out_dir, 'trajectories')
        for filename in os.listdir(trajectory_directory):
            if 'trajectory' in filename:
                filename = os.path.join(trajectory_directory, filename)
                trajectories.append(np.loadtxt(filename, skiprows=1))
            else:
                raise SimulationError("Couldn't identify file '{0}' found in \
                                        output folder".format(filename))
        if show_labels:
            with open(os.path.join(trajectory_directory, 'trajectory0.txt'), 'r') as fd:
                headers = fd.readline()
            return headers.split(), trajectories
        return trajectories


class StochKitODESolver(StochKitBaseSolver):
    name = "StochKitODESolver"
    """
    Abstract class for StochKit solver derived from the GillesPySolver class.
    This is generally used to set up the solver.


    :param model: The model on which the solver will operate
    :type model: GillesPy2.model

    :param t: The end time of the solver
    :type t: float

    :param number_of_trajectories: The number of times to sample the chemical master equation. Each
    trajectory will be returned at the end of the simulation.
    :type number_of_trajectories: int

    :param increment: The time step of the solution
    :type increment: float

    :param seed: The random seed for the simulation. Defaults to None.
    :type seed: int

    :param stochkit_home: Path to stochkit. This is set automatically upon installation, but
    may be overwritten if desired.
    :type stochkit_home: str

    :param algorithm: The solver by which to simulate the model. 'ssa' or 'tau_leaping'
    are the available options. If 'ssa' is chosen, StochKit will choose
    from the available ssa options.
    :type algorithm: str

    :param job_id: If given, this will be the name of the solver run. Usually not set.
    :type job_id: str

    :param extra_args:  Any extra arguments for the stochkit solver. See StochKit2
    documentation for details.
    :type extra_args: str

    :param debug: Set to True to provide additional debug information about the simulation.
    :type debug: bool

    :param show_labels: Use names of species as index of result object rather than position numbers.
    :type show_labels: str
    """

    @classmethod
    def run(cls, model, t=20, number_of_trajectories=1,
            increment=0.05, seed=None, stochkit_home=None,
            algorithm='stochkit_ode.py',
            job_id=None, debug=False, profile=False, show_labels=False, **kwargs):
        return super().run(model, t, number_of_trajectories, increment, seed, stochkit_home,
                           algorithm, job_id, debug=debug, show_labels=show_labels)

    @classmethod
    def get_trajectories(cls, out_dir, debug=False, show_labels=False):
        if debug:
            print("StochKitODESolver.get_trajectories(out_dir={0}".format(out_dir))
        # Collect all the output data
        trajectories = []
        with open(os.path.join(out_dir, 'output.txt'), 'r') as fd:
            fd.readline()
            headers = fd.readline()
            fd.readline()
            data = [[float(x) for x in fd.readline().split()]]
            fd.readline()
            for line in fd:
                data.append([float(x) for x in line.split()])
        trajectories.append(np.array(data))
        if show_labels:
            return headers.split(), trajectories
        return trajectories



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

"""
This file contains a function and variable for testing a machines support of GillesPy2 C++ solvers.
Used in model.py
"""

from gillespy2.solvers.cpp.build.build_engine import BuildEngine

def check_cpp_support():
    return not len(BuildEngine.get_missing_dependencies()) > 0

cpp_support = check_cpp_support()




"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import unittest
import os
import nbformat
from nbconvert.preprocessors import ExecutePreprocessor


class TestNotebooks(unittest.TestCase):
    
    def test_notebooks(self):
        FULL, MINIMAL = 0, 1
        test_set = MINIMAL
        notebooks = {}
        errors = {}
        ep = ExecutePreprocessor(timeout=600, kernel_name='python3', allow_errors=True)

        if test_set == FULL:
            test_dir = os.path.join('..', 'examples')
        elif test_set == MINIMAL:
            test_dir = os.path.join('..', os.path.join('examples', 'StartingModels'))

        for root, dirs, files in os.walk(test_dir):
            for file in files:
                if file.endswith(".ipynb"):
                     with open(os.path.join(root, file)) as f:
                        print('Reading {}...'.format(file))
                        notebooks[file] = nbformat.read(f, as_version=nbformat.NO_CONVERT)

        for file, nb in notebooks.items():
            with self.subTest(msg=file):
                try:
                    ep.preprocess(nb, {'metadata': {'path': root}})
                except Exception as err:
                    errors[file] = err
                self.assertFalse(bool(errors))


if __name__ == '__main__':
    unittest.main()




"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

"""GillesPy2 Solver for ODE solutions."""

from threading import Thread, Event
from scipy.integrate import ode
from collections import OrderedDict
import numpy as np
from gillespy2.core import GillesPySolver, log, gillespyError
from gillespy2.solvers.utilities import solverutils as nputils
from gillespy2.core.results import Results


class ODESolver(GillesPySolver):
    """
    This Solver produces the deterministic continuous solution via ODE.
    """
    name = "ODESolver"
    rc = 0
    stop_event = None
    result = None
    pause_event = None

    def __init__(self, model=None):
        name = "ODESolver"
        rc = 0
        stop_event = None
        pause_event = None
        result = None
        self.model = model

    @staticmethod
    def __f(t, y, curr_state, model, c_prop):
        """
        The right hand side of the differential equation, uses scipy.integrate ode

        :param t: time as a numpy array
        :param y: species pops as a list
        :param curr_state: dictionary of eval variables
        :param model: model being simulated
        :param c_prop: precompiled reaction propensity function

        :returns: integration step
        """
        curr_state[0]['t'] = t
        state_change =  OrderedDict()
        for i, species in enumerate(model.listOfSpecies):
            curr_state[0][species] = y[i]
            state_change[species] = 0
        propensity = OrderedDict()
        for r_name, reaction in model.listOfReactions.items():
            propensity[r_name] = eval(c_prop[r_name], curr_state[0])
            for react, stoich in reaction.reactants.items():
                state_change[react.name] -= propensity[r_name] * stoich
            for prod, stoich in reaction.products.items():
                state_change[prod.name] += propensity[r_name] * stoich
        state_change = list(state_change.values())
        return state_change

    @classmethod
    def get_solver_settings(self):
        """
        :returns: Tuple of strings, denoting all keyword argument for this solvers run() method.
        """
        return ('model', 't', 'number_of_trajectories', 'increment', 'integrator', 'integrator_options',
                'timeout')

    @classmethod
    def run(self, model=None, t=20, number_of_trajectories=1, increment=None, show_labels=True, integrator='lsoda',
            integrator_options={}, live_output=None, live_output_options={}, timeout=None, resume=None, **kwargs):
        """
        :param model: gillespy2.model class object
        :param t: end time of simulation
        :param number_of_trajectories: Should be 1.
            This is deterministic and will always have same results
        :param increment: time step increment for plotting
        :param integrator: integrator to be used form scipy.integrate.ode. Options include 'vode', 'zvode', 'lsoda',
            'dopri5', and 'dop853'.  For more details,
            see https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.ode.html

        :param integrator_options: a dictionary containing options to the scipy integrator. for a list of options,
            see https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.ode.html.
            Example use: {max_step : 0, rtol : .01}

        :param timeout: If set, if simulation takes longer than timeout, will exit.
        :type timeout: int

        :param resume: Result of a previously run simulation, to be resumed
        :param live_output: str The type of output to be displayed by solver. Can be "progress", "text", or "graph".
        :param live_output_options: dictionary contains options for live_output. By default {"interval":1}.
            "interval" specifies seconds between displaying.
            "clear_output" specifies if display should be refreshed with each displa
        """
        if isinstance(self, type):
            self = ODESolver(model=model)

        increment = self.get_increment(model=model, increment=increment)

        self.stop_event = Event()
        self.pause_event = Event()

        if timeout is not None and timeout <= 0:
            timeout = None
        if len(kwargs) > 0:
            for key in kwargs:
                log.warning('Unsupported keyword argument to {0} solver: {1}'.format(self.name, key))
        if number_of_trajectories > 1:
            log.warning("Generating duplicate trajectories for model with ODE Solver. "
                        "Consider running with only 1 trajectory.")

        if resume is not None:
            # start where we last left off if resuming a simulation
            lastT = resume['time'][-1]
            step = lastT - resume['time'][-2]
            timeline = np.arange(lastT, t+step, step)
        else:
            timeline = np.linspace(0, t, int(round(t / increment + 1)))

        species = list(model._listOfSpecies.keys())
        trajectory_base, tmpSpecies = nputils.numpy_trajectory_base_initialization(model, number_of_trajectories,
                                                                                   timeline, species, resume=resume)

        # curr_time and curr_state are list of len 1 so that __run receives reference
        if resume is not None:
            curr_time = [resume['time'][-1]]
        else:
            curr_time = [0]  # Current Simulation Time
        curr_state = [None]
        live_grapher = [None]

        sim_thread = Thread(target=self.___run, args=(model, curr_state, curr_time, timeline, trajectory_base,
                                                      tmpSpecies, live_grapher,), kwargs={'t': t,
                                                                                          'number_of_trajectories':
                                                                                              number_of_trajectories,
                                                                                          'increment': increment,
                                                                                          'timeout': timeout,
                                                                                          'resume': resume,
                                                                                          'integrator': integrator,
                                                                                          'integrator_options':
                                                                                              integrator_options, })
        try:
            time = 0
            sim_thread.start()
            if live_output is not None:
                import gillespy2.core.liveGraphing
                live_output_options['type'] = live_output

                gillespy2.core.liveGraphing.valid_graph_params(live_output_options)

                if resume is not None:
                    resumeTest = True  # If resuming, relay this information to live_grapher
                else:
                    resumeTest = False
                live_grapher[0] = gillespy2.core.liveGraphing.LiveDisplayer(model, timeline, number_of_trajectories,
                                                                            live_output_options, resume=resumeTest)
                display_timer = gillespy2.core.liveGraphing.RepeatTimer(live_output_options['interval'],
                                                                        live_grapher[0].display,
                                                                        args=(curr_state, curr_time, trajectory_base,live_output))
                display_timer.start()

            if timeout is not None:
                while sim_thread.is_alive():
                    sim_thread.join(.1)
                    time += .1
                    if time >= timeout:
                        break
            else:
                while sim_thread.is_alive():
                    sim_thread.join(.1)

            if live_grapher[0] is not None:
                display_timer.cancel()

            self.stop_event.set()
            while self.result is None:
                pass
        except KeyboardInterrupt:
            if live_output:
                display_timer.pause = True
                display_timer.cancel()
            self.pause_event.set()
            while self.result is None:
                pass
        if hasattr(self, 'has_raised_exception'):
            raise self.has_raised_exception
        
        return Results.build_from_solver_results(self, live_output_options)

    def ___run(self, model, curr_state, curr_time, timeline, trajectory_base, tmpSpecies, live_grapher, t=20,
               number_of_trajectories=1, increment=0.05, timeout=None, show_labels=True, integrator='lsoda',
               integrator_options={}, resume=None, **kwargs):
        try:
            self.__run(model, curr_state, curr_time, timeline, trajectory_base, tmpSpecies, live_grapher, t,
                       number_of_trajectories, increment, timeout, show_labels, integrator, integrator_options, resume,
                       **kwargs)
        except Exception as e:
            self.has_raised_exception = e
            self.result = []
            return [], -1

    def __run(self, model, curr_state, curr_time, timeline, trajectory_base, tmpSpecies, live_grapher, t=20,
              number_of_trajectories=1, increment=0.05, timeout=None, show_labels=True, integrator='lsoda',
              integrator_options={}, resume=None, **kwargs):

        timeStopped = 0
        if resume is not None:
            if resume[0].model != model:
                raise gillespyError.ModelError('When resuming, one must not alter the model being resumed.')
            if t < resume['time'][-1]:
                raise gillespyError.ExecutionError(
                    "'t' must be greater than previous simulations end time, or set in the run() method as the "
                    "simulations next end time")

        # compile reaction propensity functions for eval
        c_prop = OrderedDict()
        for r_name, reaction in model.listOfReactions.items():
            c_prop[r_name] = compile(reaction.ode_propensity_function, '<string>', 'eval')

        result = trajectory_base[0]
        entry_count = 0

        y0 = [0] * len(model.listOfSpecies)

        curr_state[0] = OrderedDict()

        if resume is not None:
            for i, s in enumerate(tmpSpecies):
                curr_state[0][s] = tmpSpecies[s]
                y0[i] = tmpSpecies[s]
        else:
            for i, s in enumerate(model.listOfSpecies.values()):
                curr_state[0][s.name] = s.initial_value
                y0[i] = s.initial_value

        for p_name, param in model.listOfParameters.items():
            curr_state[0][p_name] = param.value
        rhs = ode(ODESolver.__f).set_integrator(integrator, **integrator_options)
        rhs.set_initial_value(y0, curr_time[0]).set_f_params(curr_state, model, c_prop)

        while entry_count < timeline.size - 1:
            if self.stop_event.is_set():
                self.rc = 33
                break
            if self.pause_event.is_set():
                timeStopped = timeline[entry_count]
                break

            int_time = curr_time[0] + increment
            entry_count += 1
            y0 = rhs.integrate(int_time)
            curr_time[0] += increment
            for i, spec in enumerate(model.listOfSpecies):
                curr_state[0][spec] = y0[i]
                result[entry_count][i+1] = curr_state[0][spec]

        results_as_dict = {
            'time': timeline
        }
        for i, species in enumerate(model.listOfSpecies):
            results_as_dict[species] = result[:, i+1]
        results = [results_as_dict] * number_of_trajectories

        if timeStopped != 0 or resume is not None:
            results = nputils.numpy_resume(timeStopped, results, resume=resume)

        self.result = results
        return results, self.rc



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import unittest
import numpy
from gillespy2.core import Model, Species, Reaction, Parameter, RateRule
from gillespy2.core.gillespyError import *


class SimpleHybridModel(Model):
    def __init__(self, parameter_values=None):
        Model.__init__(self, name="Simple_Hybrid_Model")

        # Species
        A = Species(name='A', initial_value=0)
        B = Species(name='B', initial_value=0)
        self.add_species([A, B])

        # Parameters
        k1 = Parameter(name='k1', expression=1)
        k2 = Parameter(name='k2', expression=10)
        self.add_parameter([k1, k2])

        # Rate Rule
        rate_rule = RateRule(name='Brate', variable='B', formula="cos(t)")
        self.add_rate_rule(rate_rule)

        # Reactions
        r1 = Reaction(name='r1', reactants={A: 1}, products={}, propensity_function="k1*B")
        r2 = Reaction(name='r2', reactants={}, products={B: 1}, rate=k2)
        self.add_reaction([r1, r2])

        self.timespan(numpy.linspace(0, 1, 11))


class TestSimpleModel(unittest.TestCase):
    def setUp(self):
        self.model = SimpleHybridModel()

#    def test_this_should_fail(self):
#        name = self.model.name
#        self.assertEqual(name, "NOT Simple_Hybrid_Model", msg="Unexpected value: {}".format(name))

    def test_model_creation(self):
        name = self.model.name
        self.assertEqual(name, "Simple_Hybrid_Model", msg="Unexpected value: {}".format(name))

    def test_addingSameSpecies_ThrowsError(self):
        A = Species(name='A', initial_value=0)
        with self.assertRaises(ModelError) as ex:
            self.model.add_species(A)
        self.assertIn('Name "{}" is unavailable. A species with that name exists.'.format(A.name), str(ex.exception))

    def test_addingMultipleSameSpecies_ThrowsError(self):
        A = Species(name='A', initial_value=0)
        B = Species(name='B', initial_value=0)
        with self.assertRaises(ModelError) as ex:
            self.model.add_species([A, B])
        self.assertIn('Name "{}" is unavailable. A species with that name exists.'.format(A.name), str(ex.exception))

    def test_addingSameParameter_ThrowsError(self):
        k1 = Parameter(name='k1', expression=0)
        with self.assertRaises(ModelError) as ex:
            self.model.add_parameter(k1)
        self.assertIn('Name "{}" is unavailable. A parameter with that name exists.'.format(k1.name), str(ex.exception))

    def test_addingMultipleSameParameter_ThrowsError(self):
        k1 = Parameter(name='k1', expression=0)
        k2 = Parameter(name='k2', expression=0)
        with self.assertRaises(ModelError) as ex:
            self.model.add_parameter([k1, k2])
        self.assertIn('Name "{}" is unavailable. A parameter with that name exists.'.format(k1.name), str(ex.exception))

    def test_delete_species(self):
        self.model.delete_species('A')
        speciesList = self.model.get_all_species()
        self.assertNotIn('A', speciesList)

    def test_delete_all_species(self):
        self.model.delete_all_species()
        speciesList = self.model.get_all_species()
        self.assertNotIn('A', speciesList)
        self.assertNotIn('B', speciesList)

    def test_set_units_concentration(self):
        self.model.set_units('concentration')
        units = self.model.units
        self.assertEqual(units, 'concentration')

    def test_set_units_population(self):
        self.model.set_units('population')
        units = self.model.units
        self.assertEqual(units, 'population')

    def test_setFakeUnits_ThrowsError(self):
        with self.assertRaises(ModelError) as ex:
            self.model.set_units('nonsense')
        self.assertEqual(str(ex.exception), "units must be either concentration or population (case insensitive)")

    def test_serialization(self):
        doc = self.model.serialize()
        self.assertIsInstance(doc, str)
        self.assertIn('<Reaction>', doc)

    def test_model_has_species(self):
        species = self.model.get_species('A')
        self.assertIsInstance(species, Species, msg='{0} has incorrect type'.format(species))

    def test_add_single_species(self):
        C = Species(name='C', initial_value=0)
        self.model.add_species(C)
        species = self.model.get_species('C')
        self.assertIsInstance(species, Species, msg='{0} has incorrect type'.format(species))

    def test_model_has_species_list(self):
        speciesList = self.model.get_all_species()
        species = speciesList['A']
        self.assertIsInstance(species, Species, msg='{0} has incorrect type'.format(species))

    def test_get_parameter(self):
        parameter = self.model.get_parameter('k1')
        self.assertEqual('1', parameter.expression)
        self.assertIsInstance(parameter, Parameter, msg='{0} has incorrect type'.format(parameter))

    def test_getFakeParameter_ThrowsError(self):
        p_name = 'fake'
        with self.assertRaises(ModelError) as ex:
            parameter = self.model.get_parameter(p_name)
        self.assertEqual(str(ex.exception), "No parameter named " + p_name)

    def test_set_parameter(self):
        self.model.set_parameter('k1', '100')
        parameter = self.model.get_parameter('k1')
        self.assertEqual('100', parameter.expression)
        self.assertIsInstance(parameter, Parameter, msg='{0} has incorrect type'.format(parameter))

    def test_delete_parameter(self):
        self.model.delete_parameter('k1')
        parameterList = self.model.get_all_parameters()
        self.assertNotIn('k1', parameterList)

    def test_delete_all_parameters(self):
        self.model.delete_all_parameters()
        parameterList = self.model.get_all_parameters()
        self.assertNotIn('k1', parameterList)
        self.assertNotIn('k2', parameterList)

    def test_model_has_parameters(self):
        parameters = self.model.get_all_parameters()
        self.assertIsInstance(parameters['k1'], Parameter, msg='{0} has incorrect type'.format(parameters))
        self.assertIsInstance(parameters['k2'], Parameter, msg='{0} has incorrect type'.format(parameters))

    def test_model_parameters_correct(self):
        parameters = self.model.get_all_parameters()
        self.assertEqual(parameters['k1'].expression, '1', msg='Has incorrect expression')
        self.assertEqual(parameters['k2'].expression, '10', msg='Has incorrect expression')

    def test_model_has_rate_rules(self):
        rate_rules = self.model.listOfRateRules
        self.assertEqual(rate_rules['Brate'].variable, 'B', msg='Has incorrect species')
        self.assertEqual(rate_rules['Brate'].formula, 'cos(t)', msg='{0} has incorrect type'.format(rate_rules))

    def test_get_reaction(self):
        reaction = self.model.get_reaction('r1')
        self.assertIsInstance(reaction, Reaction, msg='{0} has incorrect type'.format(reaction))

    def test_model_has_reactions(self):
        reactions = self.model.get_all_reactions()
        self.assertIsInstance(reactions['r1'], Reaction, msg='{0} has incorrect type'.format(reactions))
        self.assertIsInstance(reactions['r2'], Reaction, msg='{0} has incorrect type'.format(reactions))

    def test_delete_reaction(self):
        self.model.delete_reaction('r1')
        reactionList = self.model.get_all_reactions()
        self.assertNotIn('r1', reactionList)

    def test_delete_all_reactions(self):
        self.model.delete_all_reactions()
        reactionList = self.model.get_all_reactions()
        self.assertNotIn('r1', reactionList)
        self.assertNotIn('r2', reactionList)

    def test_model_has_reactions_correct(self):
        reactions = self.model.get_all_reactions()

        species_A = self.model.get_species('A')
        species_B = self.model.get_species('B')

        reactants_r1 = reactions['r1'].reactants
        products_r2 = reactions['r2'].products

        species_r1 = list(reactants_r1)[0]
        species_r2 = list(products_r2)[0]

        # Check r1 name & propensity function is set
        self.assertEqual(reactions['r1'].name, 'r1', msg='Has incorrect expression')
        self.assertEqual(reactions['r1'].propensity_function, '(k1*B)', msg='Has incorrect expression')

        # Check r1 reactants are set
        self.assertEqual(reactants_r1[species_A], 1, msg='Has incorrect number of reactants')
        self.assertIsInstance(species_r1, Species, msg='Has incorrect type')

        # Check r2 products are set
        self.assertEqual(products_r2[species_B], 1, msg='Has incorrect number of products')
        self.assertIsInstance(species_r2, Species, msg='Has incorrect type')

        # Check r2 name & rate is set
        self.assertEqual(reactions['r2'].name, 'r2', msg='Has incorrect expression')
        self.assertEqual(reactions['r2'].marate.expression, '10', msg='Has incorrect expression')

    def test_model_has_timespan_correct(self):
        timespan = self.model.tspan
        self.assertCountEqual(timespan, numpy.linspace(0, 1, 11), msg='Has incorrect timespan')


if __name__ == '__main__':
    unittest.main()



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import subprocess
import signal
import threading
import queue

import numpy

from enum import IntEnum
from concurrent.futures import ThreadPoolExecutor
from typing import Union

from gillespy2.core import Model
from gillespy2.core import Results
from gillespy2.core import log
from gillespy2.core import gillespyError
from gillespy2.solvers.cpp.c_decoder import SimDecoder
from gillespy2.solvers.cpp.build.build_engine import BuildEngine
from gillespy2.solvers.cpp.build.template_gen import SanitizedModel

class SimulationReturnCode(IntEnum):
    DONE = 0
    PAUSED = 33
    FAILED = -1

class CSolver:
    """
    This class implements base behavior that will be needed for C++ solver implementees.

    :param model: The Model to simulate.
    :type model: Model

    :param output_directory: The working output directory.
    :type output_directory: str

    :param delete_directory: If True then the output_directory will be deleted upon completetion.
    :type delete_directory: bool

    :param resume: Resume data from a previous simulation run.

    :param variable: Indicates whether the simulation should be variable.
    :type variable: bool
    """
    rc = 0

    def __init__(self, model: Model = None, output_directory: str = None, delete_directory: bool = True, resume=None, variable: bool = False):
        self.delete_directory = False
        self.model = model
        self.resume = resume
        self.variable = variable
        self.build_engine: BuildEngine = None

        # Validate output_directory, ensure that it doesn't already exist
        if isinstance(output_directory, str):
            output_directory = os.path.abspath(output_directory)

            if os.path.exists(output_directory):
                raise gillespyError.DirectoryError(
                    f"Could not write to specified output directory: {output_directory}"
                    " (already exists)"
                )
        self.output_directory = output_directory
        self.delete_directory = delete_directory

        if self.model is None:
            return

        self._build(model, self.target, variable, False)
        self.species_mappings = self.model.sanitized_species_names()
        self.species = list(self.species_mappings.keys())
        self.parameter_mappings = self.model.sanitized_parameter_names()
        self.parameters = list(self.parameter_mappings.keys())
        self.reactions = list(self.model.listOfReactions.keys())
        self.result = []
        self.rc = 0

    def __del__(self):
        if self.build_engine is None:
            return

        if not self.delete_directory:
            return

        self.build_engine.clean()

    def _build(self, model: "Union[Model, SanitizedModel]", simulation_name: str, variable: bool, debug: bool = False) -> str:
        """
        Generate and build the simulation from the specified Model and solver_name into the output_dir.

        :param model: The Model to simulate.
        :type model: gillespy2.Model

        :param simulation_name: The name of the simulation to execute.
        :type simulation_name: str

        :param variable: If True the simulation will be variable, False if not.
        :type variable: bool

        :param debug: Enables or disables debug behavior.
        :type debug: bool
        """

        # Prepare the build workspace.
        if self.build_engine is None or self.build_engine.get_executable_path() is None:
            self.build_engine = BuildEngine(debug=debug, output_dir=self.output_directory)
            self.build_engine.prepare(model, variable)
            # Compile the simulation, returning the path of the executable.
            return self.build_engine.build_simulation(simulation_name)

        # Assume that the simulation has already been built.
        return self.build_engine.get_executable_path()

    def _run_async(self, sim_exec: str, sim_args: "list[str]", decoder: SimDecoder, timeout: int = 0):
        """
        Run the target executable simulation as async.

        :param sim_exec: The executable simulation to run.
        :type sim_exec: str

        :param sim_args: The arguments to pass on simulation run.
        :type sim_args: list[str]

        :param decoder: The SimDecoder instance that will handle simulation output.
        :type decoder: SimDecoder

        :returns: A future which represents the currently executing run_simulation job.
        """

        executor = ThreadPoolExecutor()
        return executor.submit(self._run, sim_exec, sim_args, decoder, timeout)

    def _run(self, sim_exec: str, sim_args: "list[str]", decoder: SimDecoder, timeout: int = 0, display_args: dict = None) -> int:
        """
        Run the target executable simulation.

        :param sim_exec: The executable simulation to run.
        :type sim_exec: str

        :param sim_args: The arguments to pass on simulation run.
        :type sim_args: list[str]

        :param decoder: The SimDecoder instance that will handle simulation output.
        :type decoder: SimDecoder

        :param display_args: The kwargs need to setup the live graphing
        :type display_args: dict

        :returns: The return_code of the simulation.
        """

        # Prefix the executable to the sim arguments.
        sim_args = [sim_exec] + sim_args

        # nt and *nix require different methods to force shutdown a running process.
        if os.name == "nt":
            proc_kill = lambda sim: sim.send_signal(signal.CTRL_BREAK_EVENT)
            platform_args = {
                "creationflags": subprocess.CREATE_NEW_PROCESS_GROUP,
                "start_new_session": True
            }

        else:
            proc_kill = lambda sim: os.killpg(sim.pid, signal.SIGINT)
            platform_args = {
                "start_new_session": True
            }

        live_grapher = [None]

        if display_args is not None:
            live_queue = queue.Queue(maxsize=1)
            def decoder_cb(curr_time, curr_state, trajectory_base=decoder.trajectories):
                try:
                    old_entry = live_queue.get_nowait()
                except queue.Empty as err:
                    pass
                curr_state = {self.species[i]: curr_state[i] for i in range(len(curr_state))}
                entry = ([curr_state], [curr_time], trajectory_base)
                live_queue.put(entry)
                
            decoder.with_callback(decoder_cb)

            from gillespy2.core.liveGraphing import (
                LiveDisplayer, CRepeatTimer, valid_graph_params
            )
            valid_graph_params(display_args['live_output_options'])
            live_grapher[0] = LiveDisplayer(**display_args)
            display_timer = CRepeatTimer(
                display_args['live_output_options']['interval'], live_grapher[0].display,
                args=(live_queue, display_args['live_output_options']['type'])
            )

        timeout_event = [False]
        with subprocess.Popen(sim_args, stdout=subprocess.PIPE, **platform_args) as simulation:
            def timeout_kill():
                timeout_event[0] = True
                proc_kill(simulation)

            timeout_thread = threading.Timer(timeout, timeout_kill)
            reader_thread = threading.Thread(target=decoder.read,
                                             args=(simulation.stdout,))

            if timeout > 0:
                timeout_thread.start()

            try:
                reader_thread.start()
                if display_args is not None:
                    display_timer.start()
                reader_thread.join()
            
            except KeyboardInterrupt:
                if live_grapher[0] is not None:
                    display_timer.pause = True
                proc_kill(simulation)

            finally:
                if live_grapher[0] is not None:
                    display_timer.cancel()
                timeout_thread.cancel()
                return_code = simulation.wait()
                reader_thread.join()

                if timeout_event[0]:
                    return SimulationReturnCode.PAUSED

                if return_code not in [0, 33]:
                    return SimulationReturnCode.FAILED

                return SimulationReturnCode.DONE

    def _make_args(self, args_dict: "dict[str, str]") -> "list[str]":
        """
        Convert a dictionary of key, value pairs into a valid Popen argument list.
        Note: Do not prefix a key with `-` as this will be handled automatically.

        :param args_dict: A dictionary of named arguments.
        :type args_dict: dict[str, str]

        :returns: A formatted list of arguments.
        """

        args_list = []

        for key, value in args_dict.items():
            args_list.extend([f"--{key}", str(value)])

        return args_list

    def _format_output(self, trajectories: numpy.ndarray):
        # Check the dimensionality of the input trajectory.
        if not len(trajectories.shape) == 3:
            raise gillespyError.ValidationError("Could not format trajectories, input numpy.ndarray is not 3-dimensional.")

        # The trajectory count is the first dimention of the input ndarray.
        trajectory_count = trajectories.shape[0]
        self.result = []

        # Begin iterating through the trajectories, copying each dimension into simulation_data.
        for trajectory in range(trajectory_count):
            # Copy the first index of the third-dimension into the simulation_data dictionary.
            data = {
                "time": trajectories[trajectory, :, 0]
            }

            for i in range(len(self.species)):
                data[self.species[i]] = trajectories[trajectory, :, i + 1]

            self.result.append(data)

        return self.result

    def _make_resume_data(self, time_stopped: int, simulation_data: numpy.ndarray, t: int):
        """
        If the simulation was paused then the output data needs to be trimmed to allow for resume.
        In the event the simulation was not paused, no data is changed.
        """
        # No need to create resume data if the simulation completed without interruption.
        # Note that, currently, some C++ solvers write 0 out as the "time stopped" by default.
        # This is likely to change in the future.
        if not time_stopped < t or time_stopped == 0:
            return simulation_data

        # Find the index of the time step value which is closest to the time stopped.
        cutoff = numpy.searchsorted(simulation_data[-1]["time"], float(time_stopped))
        if cutoff < 2:
            log.warning('You have paused the simulation too early, and no points have been calculated past'
                        ' initial values. A graphic display will not produce expected results.')

        # Break off any extraneous data which goes past the cutoff time.
        # Any data in this case is assumed to be untrusted.
        for entry_name, entry_data in simulation_data[-1].items():
            simulation_data[-1][entry_name] = entry_data[:cutoff]

        return simulation_data

    def _update_resume_data(self, resume: Results, simulation_data: "list[dict[str, numpy.ndarray]]", time_stopped: int):
        """
        Modify the simulation output to continue from a previous Results object.
        Does not handle the case where the simulation was interrupted again.
        """
        # No need to update the resume data if there is no previous data, or not enough.
        if resume is None or len(resume["time"]) < 2:
            return simulation_data

        resume_time = float(resume["time"][-1])
        increment = resume_time - float(resume["time"][-2])
        # Replace the simulation's timespan to continue where the Results object left off.
        simulation_data[-1]["time"] = numpy.arange(start=(resume_time),
                                                   stop=(resume_time + time_stopped + increment),
                                                   step=increment)

        for entry_name, entry_data in simulation_data[-1].items():
            # The results of the current simulation is treated as an "extension" of the resume data.
            # As such, the new simulation output is formed by joining the two end to end.
            new_data = numpy.concatenate((resume[entry_name], entry_data[1:]), axis=None)
            simulation_data[-1][entry_name] = new_data

        return simulation_data

    def _validate_resume(self, t: int, resume):
        """
        Validate `resume`. An exception will be thrown if resume['time'][-1] is > t.
        """

        if resume is None:
            return

        if t < resume["time"][-1]:
            raise gillespyError.ExecutionError(
                "'t' must be greater than previous simulations end time, or set in the run() method as the "
                "simulations next end time"
            )

    def _validate_kwargs(self, **kwargs):
        """
        Validate any additional kwargs passed to the model. If any exist, warn the user.
        """

        if len(kwargs) == 0:
            return

        for key, val in kwargs.items():
            log.warning(f"Unsupported keyword argument for solver {self.name}: {key}")

    def _validate_sbml_features(self, unsupported_features: "dict[str, str]"):
        detected = [ ]
        for feature_name, count in unsupported_features.items():
            if count:
                detected.append(feature_name)

        if len(detected):
            raise gillespyError.ModelError(f"Could not run Model.  SBML Feature: {detected} not supported by SSACSolver.")

    def _validate_seed(self, seed: int):
        if seed is None:
            return None

        if not isinstance(seed, int):
            seed = int(seed)

        if seed <= 0:
            raise gillespyError.ModelError("`seed` must be a postive integer.")

        return seed
        
    def _validate_variables_in_set(self, variables, values):
        for var in variables.keys():
            if var not in values:
                raise gillespyError.SimulationError(f"Argument to variable '{var}' is not a valid variable. Variables must be model species or parameters.")

    def _validate_type(self, value, typeof: type, message: str):
        if not type(value) == typeof:
            raise gillespyError.SimulationError(message)



import gillespy2
from gillespy2.solvers.cpp.c_decoder import IterativeSimDecoder
from gillespy2.solvers.utilities import solverutils as cutils
from gillespy2.core import GillesPySolver, gillespyError, Model
from typing import Union
from gillespy2.core import Results

from .c_solver import CSolver, SimulationReturnCode
from gillespy2.solvers.cpp.build.template_gen import SanitizedModel

class TauHybridCSolver(GillesPySolver, CSolver):
    name = "TauHybridCSolver"
    target = "hybrid"

    @classmethod
    def __create_options(cls, model: "SanitizedModel") -> "SanitizedModel":
        """
        Populate the given list of species modes into a set of template macro definitions.
        Generated options are specific to the Tau Hybrid solver,
        and get passed as custom definitions to the build engine.

        :param model: Sanitized model containing sanitized species definitions.
        The GPY_HYBRID_SPECIES_MODES option will be set as an option for the model.
        :type model: SanitizedModel

        :returns: Pass-through of sanitized model object.
        :rtype: SanitizedModel
        """
        species_mode_map = {
            "continuous": "CONTINUOUS_MODE",
            "discrete": "DISCRETE_MODE",
            "dynamic": "DYNAMIC_MODE",
        }
        boundary_condition_types = [
            # When species.boundary_condition == False
            "STANDARD",
            # When species.boundary_condition == True
            "BOUNDARY",
        ]

        species_mode_list = []
        for spec_id, species in enumerate(model.species.values()):
            mode_keyword = species_mode_map.get(species.mode, species_mode_map["dynamic"])
            # Casting a bool to an int evaluates: False -> 0, and True -> 1
            # Explicit cast to bool for safety, in case boundary_condition is given weird values
            boundary_keyword = boundary_condition_types[int(bool(species.boundary_condition))]
            # Example: SPECIES_MODE(2, 10, CONTINUOUS_MODE, BOUNDARY)
            entry = f"SPECIES_MODE({spec_id},{species.switch_min},{mode_keyword},{boundary_keyword})"
            species_mode_list.append(entry)

        model.options["GPY_HYBRID_SPECIES_MODES"] = " ".join(species_mode_list)
        return model

    def _build(self, model: "Union[Model, SanitizedModel]", simulation_name: str, variable: bool, debug: bool = False,
               custom_definitions=None) -> str:
        sanitized_model = TauHybridCSolver.__create_options(SanitizedModel(model))
        for rate_rule in model.listOfRateRules.values():
            sanitized_model.use_rate_rule(rate_rule)
        return super()._build(sanitized_model, simulation_name, variable, debug)

    def get_solver_settings(self):
        """
        :return: Tuple of strings, denoting all keyword argument for this solvers run() method.
        """
        return ('model', 't', 'number_of_trajectories', 'timeout', 'increment', 'seed', 'debug', 'profile')

    def run(self=None, model: Model = None, t: int = 20, number_of_trajectories: int = 1, timeout: int = 0,
            increment: int = None, seed: int = None, debug: bool = False, profile: bool = False, variables={}, 
            resume=None, live_output: str = None, live_output_options: dict = {}, tau_step: int = .03, tau_tol=0.03, **kwargs):

        if self is None or self.model is None:
            self = TauHybridCSolver(model, resume=resume)

        increment = self.get_increment(model=model, increment=increment)

        # Validate parameters prior to running the model.
        self._validate_type(variables, dict, "'variables' argument must be a dictionary.")
        self._validate_variables_in_set(variables, self.species + self.parameters)
        self._validate_resume(t, resume)
        self._validate_kwargs(**kwargs)
        self._validate_sbml_features({
            "Assignment Rules": len(model.listOfAssignmentRules),
            "Events": len(model.listOfEvents),
            "Function Definitions": len(model.listOfFunctionDefinitions)
        })

        if resume is not None:
            t = abs(t - int(resume["time"][-1]))

        number_timesteps = int(round(t / increment + 1))

        args = {
            "trajectories": number_of_trajectories,
            "timesteps": number_timesteps,
            "tau_tol": tau_tol,
            "end": t,
            "interval": str(number_timesteps),
        }

        if self.variable:
            populations = cutils.update_species_init_values(model.listOfSpecies, self.species, variables, resume)
            parameter_values = cutils.change_param_values(model.listOfParameters, self.parameters, model.volume, variables)

            args.update({
                "init_pop": populations,
                "parameters": parameter_values
            })

        seed = self._validate_seed(seed)
        if seed is not None:
            args.update({
                "seed": seed
            })

        if live_output is not None:
            live_output_options['type'] = live_output
            display_args = {
                "model": model, "number_of_trajectories": number_of_trajectories, "timeline": np.linspace(0, t, number_timesteps),
                "live_output_options": live_output_options, "resume": bool(resume)
            }
        else:
            display_args = None

        args = self._make_args(args)
        decoder = IterativeSimDecoder.create_default(number_of_trajectories, number_timesteps, len(self.model.listOfSpecies))

        sim_exec = self._build(model, self.target, self.variable, False)
        sim_status = self._run(sim_exec, args, decoder, timeout, display_args)

        if sim_status == SimulationReturnCode.FAILED:
            raise gillespyError.ExecutionError("Error encountered while running simulation C++ file:\n"
                f"Return code: {int(sim_status)}.\n")

        trajectories, time_stopped = decoder.get_output()

        simulation_data = self._format_output(trajectories)
        if sim_status == SimulationReturnCode.PAUSED:
            simulation_data = self._make_resume_data(time_stopped, simulation_data, t)
        if resume is not None:
            simulation_data = self._update_resume_data(resume, simulation_data, time_stopped)
        self.result = simulation_data
        self.rc = int(sim_status)

        return Results.build_from_solver_results(self, live_output_options)



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import ast
import json

def __add_events(model, events):
    for name, event in events.items():
        s_event = {"compID":model['defaultID'],
                   "name": name,
                   "annotation": "",
                   "delay": event.delay,
                   "priority": event.priority,
                   "triggerExpression": event.trigger.expression,
                   "initialValue": event.trigger.value,
                   "persistent": event.trigger.persistent,
                   "useValuesFromTriggerTime": event.use_values_from_trigger_time,
                   "eventAssignments": []}

        __add_event_assignments(model=model, event=s_event, assignments=event.assignments)

        model['eventsCollection'].append(s_event)
        model['defaultID'] += 1


def __add_event_assignments(model, event, assignments):
    for assignment in assignments:
        name = assignment.variable.name
        try:
            variable = __get_species(species=model['species'], name=name)
        except IndexError:
            variable = __get_parameter(parameters=model['parameters'], name=name)

        s_assignment = {"variable": variable,
                        "expression": assignment.expression}
        event['eventAssignments'].append(s_assignment)


def __add_function_definitions(model, function_definitions):
    for name, function_definition in function_definitions.items():
        variables = function_definition.get_arg_string()
        expression = function_definition.function_string
        function = "lambda({0}, {1})".format(variables, expression)
        signature = "{0}({1})".format(name, variables)

        s_function_definition = {"compID":model['defaultID'],
                                 "name":name,
                                 "function":function,
                                 "expression":expression,
                                 "variables":variables,
                                 "signature":signature,
                                 "annotation": ""}
        model['functionDefinitions'].append(s_function_definition)
        model['defaultID'] += 1


def __add_parameters(model, parameters):
    for name, parameter in parameters.items():
        try:
            expression = ast.literal_eval(parameter.expression)
        except:
            expression = parameter.expression
        s_parameter = {"compID":model['defaultID'],
                       "name":name,
                       "expression":str(expression),
                       "annotation": ""}
        model['parameters'].append(s_parameter)
        model['defaultID'] += 1


def __add_reactions(model, reactions):
    for name, reaction in reactions.items():
        s_reaction = {"compID":model['defaultID'],
                      "name":name,
                      "reactionType": "custom-propensity",
                      "massaction": False,
                      "propensity": reaction.propensity_function,
                      "annotation": "",
                      "rate": {},
                      "types": [],
                      "reactants": [],
                      "products": []}

        for key in ['reactants', 'products']:
            __add_stoich_species(s_reaction=s_reaction, reaction=reaction,
                                                key=key, species=model['species'])
        __add_summary(reaction=s_reaction)

        model['reactions'].append(s_reaction)
        model['defaultID'] += 1


def __add_rules(model, r_type, rules):
    for name, rule in rules.items():
        try:
            variable = __get_species(species=model['species'], name=rule.variable)
        except IndexError:
            variable = __get_parameter(parameters=model['parameters'], name=rule.variable)

        s_rule = {"compID":model['defaultID'],
                  "name":name,
                  "expression":rule.formula,
                  "type":r_type,
                  "variable":variable,
                  "annotation": ""}
        model['rules'].append(s_rule)
        model['defaultID'] += 1


def __add_species(model, species):
    modes = []

    for name, specie in species.items():
        if specie.mode is not None and specie.mode not in modes:
            modes.append(specie.mode)
        s_species = {"compID":model['defaultID'],
                     "name":name,
                     "value":specie.initial_value,
                     "mode":specie.mode,
                     "switchTol": specie.switch_tol,
                     "switchMin": specie.switch_min,
                     "isSwitchTol": specie.switch_min == 0,
                     "annotation": "",
                     "diffusionConst":0,
                     "types": []}
        model['species'].append(s_species)
        model['defaultID'] += 1

    if not modes:
        model['defaultMode'] = ""
    elif len(modes) > 1:
        model['defaultMode'] = "dynamic"
    else:
        model['defaultMode'] = modes[0]


def __add_stoich_species(s_reaction, reaction, key, species):
    source = reaction.reactants if key == "reactants" else reaction.products
    for specie, ratio in source.items():
        stoich_species = {"ratio":ratio,
                          "specie":__get_species(species=species, name=specie.name)}
        s_reaction[key].append(stoich_species)


def __add_summary(reaction):
    r_summary = __build_element(reaction['reactants'])
    p_summary = __build_element(reaction['products'])
    reaction['summary'] = f"{r_summary} \\rightarrow {p_summary}"


def __build_element(stoich_species):
    if not stoich_species:
        return "\\emptyset"

    elements = []
    for species in stoich_species:
        name = species['specie']['name']
        ratio = species['ratio']
        element = f"{ratio}{name}" if ratio > 1 else name
        elements.append(element)
    return '+'.join(elements)


def __get_parameter(parameters, name):
    return list(filter(lambda parameter: parameter['name'] == name, parameters))[0]


def __get_species(species, name):
    return list(filter(lambda specie: specie['name'] == name, species))[0]


def __write_to_file(model, path):
    with open(path, "w") as model_file:
        json.dump(model, model_file, indent=4, sort_keys=True)


def export(model, path=None, return_stochss_model=False):
    """
    GillesPy model to StochSS converter

    :param model: GillesPy model to be converted to StochSS
    :type model: gillespy.Model

    :param path: Path to the StochSS file for conversion
    :type path: str
    """

    if path is None:
        path = f"{model.name}.mdl"

    s_model = {"is_spatial": False,
               "defaultID": 1,
               "annotation": "",
               "volume": model.volume,
               "modelSettings": {
                    "endSim": model.tspan[-1],
                    "timeStep": model.tspan[-1] / len(model.tspan)
                },
                "species": [],
                "initialConditions": [],
                "parameters": [],
                "reactions": [],
                "rules": [],
                "eventsCollection": [],
                "functionDefinitions": []
              }

    __add_species(model=s_model, species=model.get_all_species())
    __add_parameters(model=s_model, parameters=model.get_all_parameters())
    __add_reactions(model=s_model, reactions=model.get_all_reactions())
    __add_events(model=s_model, events=model.get_all_events())
    __add_rules(model=s_model, r_type='Rate Rule', rules=model.get_all_rate_rules())
    __add_rules(model=s_model, r_type='Assignment Rule', rules=model.get_all_assignment_rules())
    __add_function_definitions(model=s_model,
                               function_definitions=model.get_all_function_definitions())

    if return_stochss_model:
        return s_model

    __write_to_file(model=s_model, path=path)
    return path



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import sys
import unittest

sys.path.append("..")
from example_models import *
from gillespy2.core import Model, Reaction, Parameter, Species, Results
from gillespy2.core.jsonify import TranslationTable

class TestJsonModels(unittest.TestCase):
    models = [
        Example,
        MichaelisMenten,
        VilarOscillator,
        Dimerization,
        Trichloroethylene,
        ToggleSwitch,
        Tyson2StateOscillator,
        Oregonator,
    ]

    runnable_models = [
        Example,
        MichaelisMenten,
        Tyson2StateOscillator,
        Schlogl
    ]

    def test_equality_of_named_models(self):
        """
        Test that a model can be converted to JSON and back.
        """

        for model in self.models:
            target = model()

            self.assertEqual(
                target, 
                model.from_json(target.to_json())
            )

    def test_equality_of_anon_models(self):
        """
        Test that an anonymous model can be converted to JSON and back.
        """

        for model in self.models:
            target = model()
            anon_target = target.to_anon()

            self.assertEqual(
                anon_target, 
                model.from_json(anon_target.to_json())
            )

    def test_equality_of_named_results(self):
        """
        Test that Model simulation results can be converted to JSON and back.
        """

        for model in self.runnable_models:
            target = model()
            results = target.run()

            results_from = Results.from_json(results.to_json())

            self.assertEquals(results.to_json(), results_from.to_json())

    def test_equality_of_named_results(self):
        """
        Test that anonymous Model simulation results can be converted to JSON and back.
        """

        for model in self.runnable_models:
            target = model()

            results = target.run()
            results._translation_table = target.get_translation_table()
            results = results.to_anon().to_json()

            self.assertEqual(results, Results.from_json(results).to_json())

    def test_model_hash_accuracy(self):
        """
        Test the accuracy of the JSON hash.
        """

        for model in self.models:
            # Create two two instances of the 'model' type.
            model_1 = model()
            model_2 = model()

            # Assert that the hash of the anonymized models are the same.
            self.assertEqual(
                model_1.to_anon().get_json_hash(), 
                model_2.to_anon().get_json_hash()
            )

            # Create a test class and change the variable insertion order.
            model_1 = model()
            model_1.var1 = "Hello"
            model_1.var2 = "world"
            model_1.var3 = [ "Hello world!" ]

            model_2 = model()
            model_2.var3 = [ "Hello world!" ]
            model_2.var2 = "world"
            model_2.var1 = "Hello"

            # Generate the first model's translation table.
            translation_table = model_1.get_translation_table()

            # Assert that the JSON of the anonymized models are still the same.
            self.assertEqual(
                model_1.to_anon().to_json(), 
                model_2.to_anon().to_json()
            )

            # Assert that the hash of the anonymized models are still the same.
            self.assertEqual(
                model_1.to_anon().get_json_hash(),
                model_2.to_anon().get_json_hash()
            )

            # Assert that the translation table is the same.
            self.assertEqual(
                model_1.get_translation_table().to_json(), 
                model_2.get_translation_table().to_json()
            )

            # Assert that model_1's JSON is equivalent to model_2 -> anon -> json -> object -> named -> json.
            self.assertEqual(
                model_1.to_json(), 
                Model.from_json(model_2.to_anon().to_json()).to_named().to_json()
            )

            # Assert that model_2's anon JSON hash is equivalent to model_2 -> anon -> json -> object -> json hash.
            self.assertEqual(
                model_1.to_anon().get_json_hash(), 
                Model.from_json(model_2.to_anon().to_json()).get_json_hash()
            )

    def test_named_to_anon_accuracy(self):
        for model in self.models:
            model_1 = model()

            # For each model, check to see if we can convert its table to and from json accurately.
            # For each model, generate its translation table and convert it to JSON.
            translation_table = model_1.get_translation_table()
            translation_table_json = translation_table.to_json()

            # Convert the JSON back into a TranslationTable object.
            translation_table_from_json = TranslationTable.from_json(translation_table_json)

            # Assert that the two tables are still identical.
            self.assertEqual(translation_table, translation_table_from_json)

            # Anonymize and convert model_1 to JSON.
            model_1 = model()
            model_1_json = model_1.to_anon().to_json()

            # Convert the JSON back into a Model object.
            model_2 = model.from_json(model_1_json)

            # Assert that the anonymized model_1 and the new model_2 are identical.
            self.assertEquals(
                model_1.to_anon().to_json(),
                model_2.to_json()
            )

            # Convert the new model_2 to named.
            model_2 = model_2.to_named()

            # Assert that model_1 and model_2 are still the same.
            self.assertEquals(
                model_1.to_json(),
                model_2.to_json()
            )

    def test_model_hash_whitespace_accuracy(self):
        """ Test that differences in whitespace do not change the hash of a model. """
        model_no_whitespace = MichaelisMenten()
        model_with_whitespace = MichaelisMenten()

         
        X = Species(name="X", initial_value=int(0.65609071 * 300.0))
        Y = Species(name="Y", initial_value=int(0.85088331 * 300.0))

        model_no_whitespace.add_species([X, Y])
        model_with_whitespace.add_species([X, Y])

        # Up to this point the JSON hash of the two models should be the same.
        self.assertEquals(model_no_whitespace.get_json_hash(), model_with_whitespace.get_json_hash())

        # Add a custom reaction to both models, differing the amount of whitespace in the 
        # propensity functions.
        reaction_no_whitespace = Reaction(
            name="X production", 
            reactants={}, 
            products={X: 1},
            propensity_function="300*1.0/(1.0+(Y*Y/(300*300)))"
        )

        reaction_with_whitespace = Reaction(
            name="X production", 
            reactants={}, 
            products={X: 1},
            propensity_function="300      * 1.0 / (1.0 + (Y  *Y         /   (300  * 300)))"
        )

        model_no_whitespace.add_reaction(reaction_no_whitespace)
        model_with_whitespace.add_reaction(reaction_with_whitespace)



import ast
from typing import Union, Optional


class Expression:
    """
    Accepts an expression string to validate and convert.
    Allows for pre-flight syntax and namespace validations,
    as well as converting between Python and C++ expressions.
    """
    def __init__(self,
                 blacklist: "Union[list[str], dict[ast.operator, str]]" = None,
                 namespace: "dict[str, any]" = None,
                 sanitize=False):
        """
        Object for managing context to validate Python expressions.
        Expressions can be passed and validated, which are validated for syntax, namespace, and other conditions.
        The later provided statements are expected to be valid Python expressions.

        :param blacklist: List of operators which are not allowed in the following expressions.
        Note that this will be "forwarded" to all following expressions.
        Ideally, one should define the "universal" blacklist in the constructor,
        using the `Expression#with_blacklist` method for more granular validations.
        :type blacklist: list[str]

        :param namespace: Dictionary mapping allowed bare identifiers to their sanitized equivalents.
        Any bare identifiers not listed as a namespace key will trigger a failed validation.
        :type  namespace: dict[str, any]

        :param sanitize: Whether or not to substitute namespace names during conversion.
        Any valid names found as namespace keys will automatically be converted to the
        corresponding namespace values in the namespace dict when getexpr_* methods are called.
        :type sanitize: bool
        """
        if blacklist is None:
            blacklist = dict({})
        elif not isinstance(blacklist, dict):
            blacklist = {ast_op: op for op, ast_op in Expression.map_operator(blacklist)}
        if namespace is None:
            namespace = {}
        self.blacklist = blacklist
        self.namespace = namespace
        self.sanitize = sanitize

    class ValidationVisitor(ast.NodeTransformer):
        def __init__(self,
                     namespace: "dict[str, str]" = None,
                     blacklist: "dict[ast.operator, str]" = None,
                     sanitize=False):
            self.namespace = dict({}) if namespace is None else namespace
            self.blacklist = dict({}) if blacklist is None else blacklist
            self.sanitize = sanitize
            self.invalid_names = []
            self.invalid_operators = []

        def check_blacklist(self, operator):
            operator = type(operator)
            if operator in self.blacklist:
                self.invalid_operators.append(str(self.blacklist.get(operator)))

        def visit_Name(self, node: "ast.Name"):
            if node.id not in self.namespace:
                self.invalid_names.append(node.id)
            elif self.sanitize:
                node.id = self.namespace.get(node.id)
            self.generic_visit(node)
            return node

        def visit_Call(self, node: "ast.Call"):
            if node.func.id not in self.namespace:
                self.invalid_names.append(node.func.id)
            elif self.sanitize:
                node.func.id = self.namespace.get(node.func.id)
            self.generic_visit(node)
            return node

        def visit_BinOp(self, node: "ast.BinOp"):
            self.check_blacklist(node.op)
            self.generic_visit(node)
            return node

        def visit_UnaryOp(self, node: "ast.UnaryOp"):
            self.check_blacklist(node.op)
            self.generic_visit(node)
            return node

        def visit_BoolOp(self, node: "ast.BoolOp"):
            self.check_blacklist(node.op)
            self.generic_visit(node)
            return node

        def visit_Compare(self, node: "ast.Compare"):
            for op in node.ops:
                self.check_blacklist(op)
            self.generic_visit(node)
            return node

        def visit_Assign(self, node: "ast.Assign"):
            self.check_blacklist(ast.Assign())
            self.generic_visit(node)
            return node

    operator_map = {
        # Basic math operators
        "+": ast.Add,
        "-": ast.Sub,
        "*": ast.Mult,
        "/": ast.Div,
        "**": ast.Pow,
        "//": ast.FloorDiv,
        "%": ast.Mod,
        "@": ast.MatMult,

        # Variable operators
        "=": ast.Assign,
        ":=": ast.Assign,

        # Boolean operators
        "and": ast.And,
        "or": ast.Or,

        # Bitwise operators (^ gets substituted for ** in Python and pow() in C++)
        "^": ast.BitXor,
        "<<": ast.LShift,
        ">>": ast.RShift,
        "|": ast.BitOr,
        "&": ast.BitAnd,

        # Comparison operators
        "==": ast.Eq,
        "!=": ast.NotEq,
        "!": ast.Not,
        ">": ast.Gt,
        ">=": ast.GtE,
        "<": ast.Lt,
        "<=": ast.LtE,
    }

    @classmethod
    def map_operator(cls, operator: "Union[str, list[str]]"):
        if isinstance(operator, list):
            for op in operator:
                yield from Expression.map_operator(op)
        else:
            # Base case: operator is a single string.
            if operator in Expression.operator_map:
                yield operator, Expression.operator_map.get(operator)
            elif operator in Expression.operator_map.values():
                # Yield the operator directly if there is no need to map it.
                yield operator

    def with_blacklist(self, blacklist: "list[str]" = None) -> "Expression":
        """
        Create a new duplicate of the current expression, with a different operator blacklist.
        Overrides operator handling behavior when converting or validating the expression.

        :param blacklist: List of operators which are not allowed.
        :type blacklist: list[ast.operator]

        :returns: List of operators which, based on the given parameters, are not valid.
        An empty list indicates that no invalid operators were found.
        """
        if blacklist is None:
            blacklist = self.blacklist.copy()
        return Expression(blacklist=blacklist, namespace=self.namespace)

    def with_namespace(self, namespace: "dict[str, any]" = None) -> "Expression":
        """
        Create a new duplicate of the current expression, with a different namespace.
        Any identifiers present in the expression which are not listed in the namespace
        will cause the expression to be flagged as an invalid namespace during validation.

        :param namespace: A dictionary containing the namespace mappings for the expression.
        The keys of the dict are expected to be the "only" valid identifiers.
        The values of the namespace are what the keys map to during sanitization, if used.
        :type namespace: dict[str, str]

        :returns: New expression containing the given namespace.
        The returned expression is a *copy* of the current expression.
        :rtype: Expression
        """
        if namespace is None:
            namespace = self.namespace.copy()
        return Expression(blacklist=self.blacklist.copy(), namespace=namespace)

    def validate(self, statement: "str") -> "ExpressionResults":
        """
        Using the information provided so far, ensure that the given Python expression is valid.
        The Python expression is parsed, raising a SyntaxError if it is an invalid Python expression.
        The expression is then checked against the given properties, such as namespace and operator blacklist.
        Additionally, the expression is rejected if it is not a single rvalue expression.

        :raises SyntaxError: The statement is not a valid Python expression.
        :returns: True if the statement is valid, otherwise returns false.
        """
        expr = ast.parse(statement)
        validator = Expression.ValidationVisitor(self.namespace, self.blacklist, self.sanitize)
        validator.visit(expr)

        return ExpressionResults(invalid_names=validator.invalid_names, invalid_operators=validator.invalid_operators)

    def __get_expr(self, converter: "ExpressionConverter") -> "Optional[str]":
        validator = Expression.ValidationVisitor(self.namespace, self.blacklist, self.sanitize)
        validator.visit(converter.tree)

        if validator.invalid_operators:
            return None

        if validator.invalid_names:
            return None

        return converter.get_str()

    def getexpr_python(self, statement: "str") -> "Optional[str]":
        """
        Converts the expression object into a Python expression string.
        Raises a SyntaxError if conversion to a Python string is impossible.

        :raises: SyntaxError
        :returns: Python expression string, if valid. Returns None if validation fails.
        """
        expr = ast.parse(statement)
        return self.__get_expr(PythonConverter(expr))

    def getexpr_cpp(self, statement: "str") -> "Optional[str]":
        """
        Converts the expression object into a C++ expression string.
        Raises a SyntaxError if conversion to a C++ string is impossible.

        :raises: SyntaxError
        :returns: C++ expression string
        """
        statement = ExpressionConverter.convert_str(statement)
        expr = ast.parse(statement)
        return self.__get_expr(CppConverter(expr))


class ExpressionResults:
    """
    Container struct for returning the results of expression validation.
    Any expression items which indicate an invalid expression are listed on an ExpressionResults instance.
    Empty lists indicate that the expression is valid.
    """
    def __init__(self, invalid_names: "list[str]" = None, invalid_operators: "list[str]" = None, is_valid=True):
        """
        Container struct for returning the results of expression validation.

        :param invalid_names: List of expression identifiers which were not valid in the given namespace.
        :type invalid_names: list[str]

        :param invalid_operators: List of blacklisted operators which were present in the expression.
        :type invalid_operators: list[str]

        :param is_valid: Override value for the `is_valid` property.
        If not set, then the validity of the expression is inferred by the `invalid_*` lists provided.
        """
        self.invalid_names = invalid_names
        self.invalid_operators = invalid_operators
        self.is_valid = is_valid and (not invalid_names and not invalid_operators)


class ExpressionConverter(ast.NodeVisitor):
    def __init__(self, tree: "ast.AST"):
        self.tree = tree
        self.expression = []

    @classmethod
    def convert_str(cls, expression: "str"):
        return expression.replace("^", "**")

    def parse_operator(self, operator: "str"):
        expr = f"({self.expression.pop()}{operator}{self.expression.pop()})"
        self.expression.append(expr)

    def parse_logical(self, operator: "str"):
        expr = f"{self.expression.pop()} {operator} {self.expression.pop()}"
        self.expression.append(expr)

    def parse_comparison(self, comparator: "str"):
        expr = f"{self.expression.pop()} {comparator} {self.expression.pop()}"
        self.expression.append(expr)

    def visit_Name(self, node: "ast.Name"):
        self.expression.append(node.id)
        self.generic_visit(node)

    def visit_Constant(self, node: "ast.Constant"):
        self.expression.append(str(node.value))
        self.generic_visit(node)

    ###########################################################################
    ### The below methods are deprecated as of Python 3.8.                  ###
    ### They are included for compatibility with 3.7 and earlier.           ###
    ### If <=3.7 becomes unsupported, please remove these visitor methods.  ###
    def visit_Num(self, node: "ast.Num"):
        self.expression.append(str(node.n))
        self.generic_visit(node)

    def visit_Str(self, node: "ast.Str"):
        self.expression.append(str(node.s))
        self.generic_visit(node)

    def visit_Bytes(self, node: "ast.Bytes"):
        self.expression.append(str(node.s))
        self.generic_visit(node)

    def visit_NameConstant(self, node: "ast.NameConstant"):
        self.expression.append(str(node.value))
        self.generic_visit(node)

    def visit_Ellipsis(self, node: "ast.Ellipsis"):
        self.expression.append(str(node))
        self.generic_visit(node)
    ### End of deprecated functions (deprecated as of Python 3.8)           ###
    ###########################################################################

    def visit_BinOp(self, node: "ast.BinOp"):
        # Right node is visited first.
        # By visiting the left node last, the most recently appended token is always the left-hand token.
        # This allows us to always append when adding to the expression, and always pop when processing it.
        self.visit(node.right)
        self.visit(node.left)
        self.visit(node.op)

    def visit_Call(self, node: "ast.Call"):
        arg_list = []
        for arg in node.args:
            self.visit(arg)
            arg_list.append(self.expression.pop())
        arg_list = ",".join(arg_list)
        expr = f"{node.func.id}({arg_list})"
        self.expression.append(expr)

    def visit_BoolOp(self, node: "ast.BoolOp"):
        # Base converter class assumes that "And" and "Or" operations are defined by inheriting class.
        # Implement visit_And() and visit_Or() to define behavior.
        for operand in reversed(node.values):
            self.visit(operand)
        # Process n-1 operations; for n operands, there are n-1 operations.
        # Example: x && y || z -> 3 operands, 2 operations
        for op_i in range(len(node.values) - 1):
            self.visit(node.op)

    def visit_Add(self, node: "ast.Add"):
        self.generic_visit(node)
        self.parse_operator("+")

    def visit_Sub(self, node: "ast.Sub"):
        self.generic_visit(node)
        self.parse_operator("-")

    def visit_Mult(self, node: "ast.Mult"):
        self.generic_visit(node)
        self.parse_operator("*")

    def visit_Div(self, node: "ast.Div"):
        self.generic_visit(node)
        self.parse_operator("/")

    def visit_Pow(self, node: "ast.Pow"):
        self.generic_visit(node)
        self.parse_operator("**")

    def visit_Compare(self, node: "ast.Compare"):
        for comparator in node.comparators:
            self.visit(comparator)
        self.visit(node.left)
        for operator in node.ops:
            self.visit(operator)

    def visit_Eq(self, node: "ast.Eq"):
        self.generic_visit(node)
        self.parse_comparison("==")

    def visit_NotEq(self, node: "ast.NotEq"):
        self.generic_visit(node)
        self.parse_comparison("!=")

    def visit_Lt(self, node: "ast.Lt"):
        self.generic_visit(node)
        self.parse_comparison("<")

    def visit_LtE(self, node: "ast.LtE"):
        self.generic_visit(node)
        self.parse_comparison("<=")

    def visit_Gt(self, node: "ast.Gt"):
        self.generic_visit(node)
        self.parse_comparison(">")

    def visit_GtE(self, node: "ast.GtE"):
        self.generic_visit(node)
        self.parse_comparison(">=")

    def visit_UnaryOp(self, node: "ast.UnaryOp"):
        self.visit(node.operand)
        self.visit(node.op)

    def visit_USub(self, node: "ast.USub"):
        self.expression.append(f"-{self.expression.pop()}")

    def _get_str(self, expr: "ast.AST"):
        self.visit(expr)
        return "".join(self.expression)

    def get_str(self) -> "str":
        return self._get_str(self.tree)


class PythonConverter(ExpressionConverter):
    def visit_And(self, node: "ast.And"):
        self.parse_logical("and")

    def visit_Or(self, node: "ast.Or"):
        self.parse_logical("or")


class CppConverter(ExpressionConverter):
    class CppExpressionTransformer(ast.NodeTransformer):
        def visit_BinOp(self, node: "ast.BinOp"):
            self.generic_visit(node)
            if isinstance(node.op, ast.Pow):
                node = ast.copy_location(ast.Call(
                    func=ast.Name(id='pow', ctx=ast.Load()),
                    args=[node.left, node.right],
                    keywords=[]
                ), node)
            return node

    def visit_And(self, node: "ast.And"):
        self.generic_visit(node)
        self.parse_logical("&&")

    def visit_Or(self, node: "ast.Or"):
        self.generic_visit(node)
        self.parse_logical("||")

    def get_str(self) -> "str":
        expr = CppConverter.CppExpressionTransformer().visit(self.tree)
        return super()._get_str(expr)



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import unittest
import os
import tempfile
from gillespy2.core import Model
from gillespy2.core.results import Results, Trajectory

class TestResults(unittest.TestCase):

    def test_xscale_plot(self):
        from unittest import mock
        trajectory = Trajectory(data={'time':[0.],'foo':[1.]}, model=Model('test_model'))
        results = Results(data=[trajectory])
        with mock.patch('matplotlib.pyplot.xscale') as mock_xscale:
            results.plot(xscale='log')
        mock_xscale.assert_called_with('log')

    def test_yscale_plot(self):
        from unittest import mock
        trajectory = Trajectory(data={'time':[0.],'foo':[1.]}, model=Model('test_model'))
        results = Results(data=[trajectory])
        with mock.patch('matplotlib.pyplot.yscale') as mock_yscale:
            results.plot(yscale='log')
        mock_yscale.assert_called_with('log')

    def test_xscale_plot_std_dev_range(self):
        from unittest import mock
        trajectory = Trajectory(data={'time':[0.],'foo':[1.]}, model=Model('test_model'))
        results = Results(data=[trajectory])
        with mock.patch('matplotlib.pyplot.xscale') as mock_xscale:
            results.plot_std_dev_range(xscale='log')
        mock_xscale.assert_called_with('log')

    def test_yscale_plot_std_dev_range(self):
        from unittest import mock
        trajectory = Trajectory(data={'time':[0.],'foo':[1.]}, model=Model('test_model'))
        results = Results(data=[trajectory])
        with mock.patch('matplotlib.pyplot.yscale') as mock_yscale:
            results.plot(yscale='log')
        mock_yscale.assert_called_with('log')

    def test_plotly_layout_args(self):
        from unittest import mock
        trajectory = Trajectory(data={'time':[0.],'foo':[1.]}, model=Model('test_model'))
        results = Results(data=[trajectory])
        with mock.patch('plotly.offline.init_notebook_mode') as mock_notebook:
            with mock.patch('plotly.graph_objs.Scatter') as mock_scatter:
                with mock.patch('plotly.graph_objs.Layout') as mock_layout:
                    results.plotplotly(return_plotly_figure=True,xscale='log')
        mock_layout.assert_called_with(showlegend=True, title='', xaxis_title='Time', xscale='log', yaxis_title='Value')

    def test_plotly_std_dev_range_layout_args(self):
        from unittest import mock
        from unittest.mock import MagicMock
        trajectory = Trajectory(data={'time':[0.],'foo':[1.]}, model=Model('test_model'))
        results = Results(data=[trajectory])
        _plotly_iterate=MagicMock()
        with mock.patch('plotly.offline.init_notebook_mode') as mock_notebook:
            with mock.patch('plotly.graph_objs.Scatter') as mock_scatter:
                with mock.patch('plotly.graph_objs.Layout') as mock_layout:
                    results.plotplotly_std_dev_range(return_plotly_figure=True, xscale='log')
        mock_layout.assert_called_with(legend={'traceorder':'normal'},showlegend=True, title='Mean and Standard Deviation',
                                       xaxis_title='Time', xscale='log', yaxis_title='Value')

    def test_pickle_stable_plot_iterate(self):
        from unittest import mock
        from gillespy2.core.results import _plot_iterate
        trajectory = Trajectory(data={'time':[0.],'foo':[1.]}, model=Model('test_model'))
        import pickle
        trajectory_unpickled = pickle.loads(pickle.dumps(trajectory))
        import matplotlib
        with mock.patch('matplotlib.pyplot.plot') as mock_method_before_pickle:
            _plot_iterate(trajectory)
        with mock.patch('matplotlib.pyplot.plot') as mock_method_after_pickle:
            _plot_iterate(trajectory_unpickled)
        assert mock_method_before_pickle.call_args_list == mock_method_after_pickle.call_args_list

    def test_pickle_stable_plotplotly_iterate(self):
        from unittest import mock
        from gillespy2.core.results import _plotplotly_iterate
        trajectory = Trajectory(data={'time':[0.],'foo':[1.]}, model=Model('test_model'))
        import pickle
        trajectory_unpickled = pickle.loads(pickle.dumps(trajectory))
        with mock.patch('plotly.graph_objs.Scatter') as mock_method_before_pickle:
            _plotplotly_iterate(trajectory)
        with mock.patch('plotly.graph_objs.Scatter') as mock_method_after_pickle:
            _plotplotly_iterate(trajectory_unpickled)
        assert mock_method_before_pickle.call_args_list == mock_method_after_pickle.call_args_list

    def test_to_csv_single_result_no_data(self):
        result = Results(data=None)
        test_nametag = "test_nametag"
        test_stamp = "test_stamp"

        with tempfile.TemporaryDirectory() as tempdir:
            result.to_csv(stamp = test_stamp, nametag = test_nametag, path=tempdir)
            test_path = tempdir+"/"+test_nametag+test_stamp
            assert not os.path.isdir(test_path)
            
    def test_to_csv_single_result_directory_exists(self):
        test_data = {'time':[0]}
        result = Results(data=[test_data])
        test_nametag = "test_nametag"
        test_stamp = "test_stamp"

        with tempfile.TemporaryDirectory() as tempdir:
            result.to_csv(stamp = test_stamp, nametag = test_nametag, path=tempdir)
            test_path = tempdir+"/"+test_nametag+test_stamp
            assert os.path.isdir(test_path)

    def test_to_csv_single_result_file_exists(self):
        test_data = {'time':[0]}
        result = Results(data=[test_data])
        test_nametag = "test_nametag"
        test_stamp = "test_stamp"

        with tempfile.TemporaryDirectory() as tempdir:
            result.to_csv(stamp = test_stamp, nametag = test_nametag, path=tempdir)
            test_path = tempdir+"/"+test_nametag+test_stamp+"/"+test_nametag+"0.csv"
            assert os.path.isfile(test_path)

    def test_to_csv_single_result_no_stamp(self):
        test_data = {'time':[0]}
        result = Results(data=[test_data])
        test_nametag = "test_nametag"

        with tempfile.TemporaryDirectory() as tempdir:
            result.to_csv(nametag=test_nametag, path=tempdir)
            assert len(os.listdir(tempdir)) is not 0

    def test_to_csv_single_result_no_nametag(self):
        test_model = Model('test_model')
        test_data = Trajectory(data={'time':[0]},model=test_model)

        with tempfile.TemporaryDirectory() as tempdir:
            result.to_csv(nametag=test_nametag, path=tempdir)
            assert len(os.listdir(tempdir)) is not 0

    def test_to_csv_single_result_no_nametag(self):
        test_model = Model('test_model')
        test_data = Trajectory(data={'time':[0]},model=test_model)
        result = Results(data=[test_data])
        result.solver_name = 'test_solver'
        test_stamp = "test_stamp"

        with tempfile.TemporaryDirectory() as tempdir:
            result.to_csv(stamp=test_stamp, path=tempdir)
            assert len(os.listdir(tempdir)) is not 0

    def test_to_csv_single_result_no_path(self):
        test_data = Trajectory({'time':[0]},model=Model('test_model'),solver_name='test_solver_name')
        result = Results(data=[test_data])
        test_nametag = "test_nametag"
        test_stamp = "test_stamp"




"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from gillespy2.core.sortableobject import SortableObject
from gillespy2.core.jsonify import Jsonify

class FunctionDefinition(SortableObject, Jsonify):
    """
    Object representation defining an evaluable function to be used during
    simulation of a GillesPy2 model

    :param name: Name of the function to be made and called
    :type name: str

    :param function: Defined function body of operation to be performed.
    :type function: str

    :param variables: String names of Variables to be used as arguments to function.
    :type variables: list[str]
    """

    def __init__(self, name="", function=None, args=[]):
        if function is None:
            raise TypeError("Function string provided for FunctionDefinition cannot be None")

        self.name = name
        self.function_string = function
        self.args = args


    def __str__(self):
        return f"{self.name}: Args: {self.args}, Expression: {self.function_string}"

    def get_arg_string(self) -> str:
        """
        Convert function's argument list into a comma-separated formatted string.

        :returns: Argument list as a comma-separated formatted string.
        """
        return ','.join(self.args)

    def sanitized_function(self, species_mappings, parameter_mappings):
        names = sorted(list(species_mappings.keys()) + list(parameter_mappings.keys()), key=lambda x: len(x),
                       reverse=True)
        replacements = [parameter_mappings[name] if name in parameter_mappings else species_mappings[name]
                        for name in names]
        sanitized_function = self.function_string
        for id, name in enumerate(names):
            sanitized_function = sanitized_function.replace(name, "{" + str(id) + "}")
        return sanitized_function.format(*replacements)


"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

"""
Test to determine if the behavior exhibited by the new check_cpp_support() function matches
what is expected by the previous try/catch implementation.
"""

import unittest
from gillespy2.solvers.cpp.build.build_engine import BuildEngine

class TestCheckCPPSupport(unittest.TestCase):
    def test_check_cpp_support(self):
        self.assertEqual(not len(BuildEngine.get_missing_dependencies()), self.old_check_cpp_support())

    def old_check_cpp_support(self):
        from gillespy2.solvers.cpp.example_models import Example
        from gillespy2 import SSACSolver
        try:
            model = Example()
            results = model.run(solver=SSACSolver, cpp_support=True)
            return True
        except Exception as e:
            from gillespy2.core import log
            log.warn('Unable to use C++ optimized SSA: {0}.  The performance of ' \
            'this package can be significantly increased if you install/configure GCC on ' \
            'this machine.'.format(e))
            return False

if __name__ == '__main__':
    unittest.main()



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import shutil
import tempfile
from pathlib import Path
from typing import Union

import gillespy2
from gillespy2.core import Model, gillespyError

from . import template_gen
from .make import Make


class BuildEngine():
    template_definitions_name = "template_definitions.h"
    template_options_name = "template_opts.h"

    def __init__(self, debug: bool = False, output_dir: str = None):
        self.self_dir = Path(__file__).parent
        self.cpp_dir = self.self_dir.joinpath("../c_base").resolve()
        self.makefile = self.cpp_dir.joinpath("Makefile")
        self.src_template_dir = self.cpp_dir.joinpath("template")
        self.output_dir = output_dir

        self.debug = debug

        # The output_dir will be generated if it does not exist on prepare().
        # Output files are all rooted relative to the output_dir.
        if self.output_dir is not None:
            self.output_dir = Path(output_dir)

            if not self.output_dir.is_dir():
                self.output_dir.mkdir(parents=True)

    def __get_cache_dir(self) -> Path:
        cache_dir = gillespy2._global_cache

        if not cache_dir.is_dir() or not os.access(str(cache_dir), os.R_OK):
            cache_dir = Path(os.path.expanduser("~"), ".gillespy2", "cache")

        return cache_dir

    @classmethod
    def get_missing_dependencies(cls) -> "list[str]":
        """
        Determine which dependencies are missing on the system, if any.

        :returns: A list of missing dependencies.
        """

        dependencies = ["g++", "make"]
        missing = [(dep) for dep in dependencies if shutil.which(dep) is None]

        return missing

    def prepare(self, model: "Union[Model, template_gen.SanitizedModel]", variable=False) -> str:
        """
        Prepare the template directory for compilation.
        The following operations will be performed:
            1. If no cached object files are found, prebuild them.
            2. Copy the C++ template directory into a new temp directory.
            3. Remove the sample template_definitions.h file.
            4. Generate and write a template_definitions.h file from the model.

        :param model: A GillesPy2 model who's template definitions will be generated.
        :type model: gillespy2.Model

        :param variable: A template_gen argument requirement which enables support for non-constant param values.
        :type variable: bool

        :returns: The path of the output directory.
        """

        # If the output directory is None, create and set it to a temporary directory.
        if self.output_dir is None:
            self.output_dir = Path(tempfile.mkdtemp())

        # If object files haven't been compiled yet, go ahead and compile them with make.
        # Precompilation only happens if the cache is enabled but hasn't been built yet.
        # Make target for individual simulation will still succeed if prebuild() isn't called.
        self.obj_dir = self.output_dir.joinpath("obj")
        self.template_dir = self.output_dir.joinpath("template")

        if gillespy2.cache_enabled:
            self.obj_dir = self.__get_cache_dir()

        if not self.obj_dir.is_dir():
            self.obj_dir.mkdir(parents=True)

        # Copy the C++ template directory to the temp directory.
        shutil.copytree(self.src_template_dir, self.template_dir)

        # If a raw GillesPy2 model was provided, convert it to a sanitized model.
        if isinstance(model, gillespy2.Model):
            model = template_gen.SanitizedModel(model)
        elif not isinstance(model, template_gen.SanitizedModel):
            raise TypeError(f"Build engine expected gillespy2.Model or SanitizedModel type: received {type(model)}")

        # Build the template and write it to the temp directory and remove the sample template_definitions header.
        template_file = self.template_dir.joinpath(self.template_definitions_name)
        template_file.unlink()
        template_gen.write_definitions(str(template_file), model.get_template(variable=variable))
        custom_definitions = model.get_options()
        if custom_definitions is not None:
            options_file = self.template_dir.joinpath(self.template_options_name)
            options_file.unlink()
            template_gen.write_definitions(str(options_file), custom_definitions)

        # With all required information gathered, create a Make instance.
        self.make = Make(str(self.makefile), str(self.output_dir), str(self.obj_dir))

        return self.output_dir

    def build_cache(self, cache_dir: str, force_rebuild: bool = False):
        """
        Build object dependencies and cache into directory for later use.

        :param cache_dir: The directory to use as a cache.
        :type cache_dir: str

        :param force_rebuild: Delete and rebuild the cache directory.
        :type force_rebuild: bool
        """

        make = Make(self.makefile, cache_dir, cache_dir)
        make.prebuild()

    def build_simulation(self, simulation_name: str) -> str:
        """
        Build the solver to the temp directory.

        :param simulation_name: The name of the simulation to build. For example, ODESimulation.
        :type simulation_name: str

        :return: The path of the newly build solver executable.
        """

        if self.make is None:
            raise gillespyError.BuildError(
                "Failed to build the simulation. The build environment has not yet been prepared.\n"
                "To fix, call `BuildEngine.prepare()` prior to attempting to build the simulation."
            )

        self.make.build_simulation(simulation_name, template_dir=str(self.template_dir))
        return str(self.make.output_file)

    def get_executable_path(self) -> str:
        """
        Resolves the filepath of the simulation executable.
        Only valid after the simulation has been built.

        :return: String containing path to executable.
            None if no executable exists.
        """
        if not os.path.exists(self.make.output_file):
            return None
        return str(self.make.output_file)

    def clean(self):
        """
        Delete the output directory and all other associated build artifacts.
        """

        if self.debug:
            return

        if self.output_dir is None:
            return

        if self.output_dir.exists():
            shutil.rmtree(self.output_dir)


"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import re
import copy
import json
import pydoc
import hashlib

from json import JSONEncoder

import numpy
import gillespy2

class Jsonify:
    """
    Interface to allow for instances of arbitrary types to be encoded into json strings
    and decoded into new objects.
    """

    _translation_table = None

    def to_json(self, encode_private=True) -> str:
        """
        Convert self into a json string.

        :param encode_private: If True all private (prefixed of '_') will be encoded. None if False.
        :type encode_private: bool

        :returns: The JSON representation of self.
        """

        encoder = ComplexJsonCoder(encode_private=encode_private)
        return json.dumps(copy.deepcopy(self), indent=4, sort_keys=True, default=encoder.default)

    @classmethod
    def from_json(cls, json_str: str) -> object:
        """
        Convert some json_str into a decoded Python type. This function should
        return a new instance of the type.

        :param json_str: A json str to be converted into a new type instance.
        :type json_str: str

        :returns: A decoded object.
        """

        # If the json_str is actually a dict, it means we've decoded as much as possible.
        if isinstance(json_str, dict):
            return cls.from_dict(json_str)

        decoder = ComplexJsonCoder()
        return json.loads(json_str, object_hook=decoder.decode)

    def to_dict(self) -> dict:
        """
        Convert the object into a dictionary ready for json encoding.
        Note: Complex types that inherit from Jsonify do not need to be manually encoded.

        By default, this function will return a dictionary of the object's public types.

        :returns: The backing var dictionary of the object.
        """

        return self.__dict__.copy()

    @classmethod
    def from_dict(cls, src_dict: dict) -> object:
        """
        Convert some dict into a new instance of a python type.
        This function will return a __new__ instance of the type.

        :param src_dict: The dictionary to apply onto the new instance.
        :type src_dict: dict

        :returns: A new object with its backing __dict__ set to a copy of src_dict.
        """

        new = cls.__new__(cls)
        new.__dict__ = src_dict.copy()

        return new

    def to_anon(self):
        """
        Converts self into an anonymous instance of self.
        """

        return self.get_translation_table().obj_to_anon(copy.deepcopy(self))

    def to_named(self):
        """
        Converts self into a named instance of self.
        """

        return self.get_translation_table().obj_to_named(copy.deepcopy(self))

    def get_translation_table(self) -> "TranslationTable":
        """
        Make and/or return the translation table.

        :returns: A TranslationTable instance.
        """

        if self._translation_table is None:
            self._translation_table = self.make_translation_table()

        return self._translation_table

    def make_translation_table(self) -> "TranslationTable":
        """
        Make a translation table that describes key:value pairs to convert
        user-define data into generic equivalents.

        :returns: A newly generated TranslationTable instance.
        """
        raise NotImplementedError("make_translation_table() has not been implemented.")

    def public_vars(self) -> dict:
        """
        Gets a dictionary of public vars that exist on self. Keys starting with '_' are ignored.

        :returns: A dictionary containing all public ('_'-prefixed) variables on the object.
        """
        return {k: v for k, v in vars(self).items() if not k.startswith("_")}

    def get_json_hash(self, ignore_whitespace=True, hash_private_vals=False) -> str:
        """
        Get the hash of the json representation of self.

        :param ignore_whitespace: If set to True all whitespace will be stripped from the JSON
            prior to being hashed.
        :type ignore_whitespace: bool

        :param hash_private_vals: If set to True all private and non-private variables will 
            be included in the hash.
        :type hash_private_vals: bool

        :returns: An MD5 hash of the object's sorted JSON representation.
        """

        # If _hash_private_vars is set, hash ALL properties on the object.
        model_json = self.to_json(encode_private=hash_private_vals)

        # If ignore_whitespace is set, strip out all whitespace characters.
        if ignore_whitespace:
            model_json = re.sub(r"\s+", "", model_json)

        return hashlib.md5(str.encode(model_json)).hexdigest()

    def __eq__(self, o: "Jsonify"):
        """
        Overload to compare the json of two objects that derive from Jsonify.
        This method will not do any additional translation.

        :param o: The Jsonify object to compare against.
        :type o: Jsonify

        :returns: True if equal, False if not.
        """
        return self.get_json_hash() == o.get_json_hash()

class ComplexJsonCoder(JSONEncoder):
    """
    This class delegates the encoding and decoding of objects to one or more implementees.

    :param translation_table: A TranslationTable instance that will be used to
        translate objects.
    :type translation_table: TranslationTable

    :param encode_private: If set to True then all private and public variables will
        be converted to JSON. If False, only public.
    :type encode_private: bool
    """

    def __init__(self, translation_table=None, encode_private=True, **kwargs):
        super().__init__(**kwargs)
        self._translation_table = translation_table
        self._encode_private = encode_private

        self._delegation_table = {
            numpy.ndarray: NdArrayCoder,
            set: SetCoder,
            type: TypeCoder
        }

    def default(self, o: object):
        """
        This function is called when json.dumps() fires. default() is a bad name for the function,
        but anything else makes JSONEncoder freak out.

        :param o: The object that is currently being encoded into JSON.
        """

        # If o is of matching type, use a custom coder.
        for obj_type, coder in self._delegation_table.items():
            if isinstance(o, obj_type):
                return coder.to_dict(o)

        if not isinstance(o, Jsonify):
            return super().default(o)

        if self._encode_private:
            model = o.to_dict()

        else:
            # Strip private variables from the object.
            model = {}

            for key, val in o.to_dict().items():
                if key.startswith("_") and not key.startswith("__"):
                    continue

                model[key] = val

        # If the model is some subclass of gillespy2.core.model.Model, then manually set its type.
        if issubclass(o.__class__, gillespy2.core.Model):
            model["_type"] = f"{gillespy2.core.Model.__module__}.{gillespy2.core.Model.__name__}"

        else:
            model["_type"] = f"{o.__class__.__module__}.{o.__class__.__name__}"

        return model

    def decode(self, json_dict: dict):
        """
        Decode the JSON dictionary into a valid Python object.

        :param json_dict: The JSON dictionary to decode.
        :type json_dict: dict

        :returns: The decoded form of the JSON dictionary.
        """

        # _type is a field embedded by the encoder to indicate which
        # Jsonify instance will be used to decode the json string.
        if "_type" not in json_dict:
            return json_dict

        json_type = pydoc.locate(json_dict["_type"])

        if json_type is None:
            raise Exception(f"{json_type} does not exist.")

        # If the type is not a subclass of Jsonify, throw an exception.
        # We do this to prevent the execution of arbitrary code.
        if not issubclass(json_type, Jsonify):
            raise Exception(f"{json_type}")

        return json_type.from_json(json_dict)

class TranslationTable(Jsonify):
    """
    This class contains functions to enable arbitrary object trees to be "translated" to anonymous
    and named objects. This behavior is defined by a map of 'named' and 'anon' key values.

    :param to_anon: A mapping of 'named' to 'anonymous' strings to be used when converting
        user-defined names to anon.
    :type to_anon: dict[str, str]
    """

    def __init__(self, to_anon: "dict[str, str]"):
        self.to_anon = to_anon.copy()
        self.to_named = dict((v, k) for k, v in list(self.to_anon.items()))

    def obj_to_anon(self, obj: object):
        """
        Recursively anonymise all named properties on the object.

        :param obj: The object to anonymize.
        :type obj: object

        :returns: An anonymized instance of self.
        """

        return self.recursive_translate(obj, self.to_anon)

    def obj_to_named(self, obj):
        """
        Recursively identify all anonymous properties on the object.

        :param obj: The object that will be converted to named.
        :type obj: object

        :returns: A named instance of self.
        """

        return self.recursive_translate(obj, self.to_named)

    def recursive_translate(self, obj: object, translation_table: "dict[str, str]"):
        """
        Recursively search through the object tree searching for property value matches in the
        translation table. If a match is found, substitute.

        :param obj: The object that will be translated.
        :type obj: object

        :param translation_table: The mapping to translate by.
        :type translation_table: TranslationTable
        """

        # If a translation table exists on the object, remove and save it.
        if "_translation_table" in obj.__dict__:
            saved_table = obj.__dict__.pop("_translation_table")

        translated = self._recursive_translate(obj, translation_table)

        # Restore the original translation table, if needed.
        if saved_table is not None:
            obj.__dict__["_translation_table"] = saved_table

        return translated

    def _recursive_translate(self, obj: object, translation_table: "dict[str, str]"):
        # The obj is a class if it's an instance of Jsonify. Class property names *cannot*
        # be changed, so translate just the values.
        if isinstance(obj, Jsonify):
            for key in vars(obj).keys():
                vars(obj)[key] = self._recursive_translate(vars(obj)[key], translation_table)

        elif isinstance(obj, list):
            for item in obj:
                item = self._recursive_translate(item, translation_table)

        elif isinstance(obj, dict):
            # Convert the dictionary into a list of tuples.
            # This makes it easier to modify key names.
            obj = list((k, v) for k, v in obj.items())
            new_pairs = [ ]

            for pair in obj:
                new_pairs.append((
                    self._recursive_translate(pair[0], translation_table),
                    self._recursive_translate(pair[1], translation_table)
                ))

            obj = dict((x[0], x[1]) for x in new_pairs)

        # If the obj is a string, translate it via a regex replace.
        # Note: mathematical functions contain additional characters that should not be translated.
        elif isinstance(obj, str):
            # To handle functions, grab all complete words from the string.
            matches = re.finditer("([0-z])+", obj)

            # For each match, translate the group.
            for match in matches:
                group = match.group()
                obj = obj.replace(group, translation_table.get(group, group))

        return obj

class NdArrayCoder(Jsonify):
    """ This JSON coder enables support for  the `numpy.ndarray` type. """

    @staticmethod
    def to_dict(obj):
        return {
            "data": obj.tolist(),
            "_type": f"{NdArrayCoder.__module__}.{NdArrayCoder.__name__}"
        }

    @staticmethod
    def from_json(obj):
        return numpy.array(obj["data"])

class SetCoder(Jsonify):
    """ This JSON coder enables support for the `set` type. """

    @staticmethod
    def to_dict(obj):
        return {
            "data": list(obj),
            "_type": f"{SetCoder.__module__}.{SetCoder.__name__}"
        }

    @staticmethod
    def from_json(obj):
        return set(obj["data"])

class TypeCoder(Jsonify):
    """ This JSON coder enables support for the 'type' type. """

    @staticmethod
    def to_dict(obj):
        return {
            "data": type(obj),
            "_type": f"{TypeCoder.__module__}.{TypeCoder.__name__}"
        }

    @staticmethod
    def from_json(obj):
        return pydoc.locate(obj["data"])



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import unittest
from example_models import *
from gillespy2.core.gillespyError import *
from gillespy2 import ODESolver
#from TestBattery import *

class TestExampleModels(unittest.TestCase):

    def test_trichloroethylene_example(self):
        trichloroethylene_model = Trichloroethylene()
        results = trichloroethylene_model.run()

    def test_lacOperon_example(self):
        lacOperon_model = LacOperon()
        results = lacOperon_model.run(solver=ODESolver)

    def test_schlogl_example(self):
        schlogl_model = Schlogl()
        results = schlogl_model.run()

    def test_michaelisMenten_example(self):
        michaelisMenten_model = MichaelisMenten()
        results = michaelisMenten_model.run()

    def test_toggleSwitch_example(self):
        toggleSwitch_model = ToggleSwitch()
        results = toggleSwitch_model.run()

    def test_example_example(self):
        example_model = Example()
        results = example_model.run()

    def test_tyson2StateOscillator_example(self):
        tyson2StateOscillator_model = Tyson2StateOscillator()
        results = tyson2StateOscillator_model.run()
    
    #def test_test_battery(self):
        #timing_battery(100, 0.1)

if __name__ == '__main__':
    unittest.main()



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# =============================================================================
# @file    __version__.py
# @brief   GillesPy2 version info
# @license Please see the file named LICENSE in the project directory
# @website https://github.com/StochSS/GillesPy2
# =============================================================================


__version__      = '1.6.6'

__title__        = 'GillesPy2'
__description__  = 'Python interface for Gillespie-style biochemical simulations'
__url__          = 'https://github.com/StochSS/GillesPy2'
__download_url__ = 'https://pypi.org/project/gillespy2/#files'
__author__       = 'See AUTHORS'
__email__        = 'bdrawert@unca.edu'
__license__      = 'GPL'
__copyright__    = 'Copyright (C) 2017-2021'



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import unittest
import numpy as np
from example_models import Example
from gillespy2 import TauLeapingSolver


class TestBasicTauLeapingSolver(unittest.TestCase):
    model = Example()
    

if __name__ == '__main__':
    unittest.main()



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from collections import OrderedDict
from typing import Optional
from gillespy2.core import Species, Reaction, Parameter, Model, RateRule
from gillespy2.solvers.cpp.build.expression import Expression
from gillespy2.core import log
import math


class SanitizedModel:
    """
    Utility class for preparing a model to be passed to a C++ solver.
    Ensures that any sanitized mappings are consistent between different mappings.
    Wraps around an existing GillesPy2 model to provide sanitized mappings.

    :param model: GillesPy2 model to produce sanitized mappings of.
    :type model: gillespy2.Model
    """
    reserved_names = {
        "vol": "V",
        "t": "t",
    }

    def __init__(self, model: Model):
        self.model = model

        self.species: "OrderedDict[str, Species]" = OrderedDict()
        self.species_names = model.sanitized_species_names()
        for species_name, sanitized_name in self.species_names.items():
            self.species[sanitized_name] = model.get_species(species_name)

        self.parameters: "OrderedDict[str, Parameter]" = OrderedDict()
        self.parameter_names = model.sanitized_parameter_names()
        for parameter_name, sanitized_name in self.parameter_names.items():
            self.parameters[sanitized_name] = model.get_parameter(parameter_name) \
                if parameter_name != "vol" else Parameter(name="V", expression=model.volume)

        base_namespace = {
            # ORDER IS IMPORTANT HERE!
            # All "system" namespace entries should always be first.
            # Otherwise, user-defined identifiers (like, for example, "gamma") might get overwritten.
            **{name: name for name in math.__dict__.keys()},
            **self.species_names,
            **self.parameter_names,
            **self.reserved_names,
        }
        self.expr = Expression(namespace=base_namespace, blacklist=["="], sanitize=True)

        # SSA Propensities: Maps reaction names to their corresponding propensity function.
        self.propensities: "OrderedDict[str, str]" = OrderedDict()
        # ODE Propensities: Maps reaction names to their corresponding mass-action rate expression.
        self.ode_propensities: "OrderedDict[str, str]" = OrderedDict()
        # Reactions: maps reaction names to their stoichiometry matrix.
        # Stoichiometry matrix maps a sanitized species name to its stoichiometry.
        self.reactions: "OrderedDict[str, dict[str, int]]" = OrderedDict()
        # Rate Rules: maps sanitized species names to their corresponding rate rule expression.
        self.rate_rules: "OrderedDict[str, str]" = OrderedDict()
        # Options: custom definitions that can be supplied by the solver, maps macros to their definitions.
        # The solver itself may use `options` to supply their own solver-specific definitions.
        self.options: "OrderedDict[str, str]" = OrderedDict()

        for reaction in model.get_all_reactions().values():
            self.use_reaction(reaction)
            self.use_propensity(reaction, ode=False)
            self.use_propensity(reaction, ode=True)

    def use_propensity(self, reaction: "Reaction", ode=False):
        """
        Populates the given reaction's propensities into the sanitized model.
        Expression conversion and sanitization are automatically applied.

        :param reaction: Reaction to generate propensities from.
        :type reaction: gillespy2.Reaction

        :param ode: Determines whether the stochastic propensity or ODE mass-action rate is used.
        If set to true, the mass-action rate is used instead of the propensity.
        """
        propensities = self.ode_propensities if ode else self.propensities
        propensity = reaction.ode_propensity_function if ode else reaction.propensity_function
        propensities[reaction.name] = self.expr.getexpr_cpp(propensity)

    def use_reaction(self, reaction: "Reaction") -> "SanitizedModel":
        """
        Adds the given reaction to the sanitized model.
        Populates the name and stoichiometry matrix into the model.

        :param reaction: Reaction to add to the sanitized model.
        :type reaction: gillespy2.Reaction
        """
        self.reactions[reaction.name] = {spec: int(0) for spec in self.species_names.values()}
        for reactant, stoich_value in reaction.reactants.items():
            if isinstance(reactant, Species):
                reactant = self.species_names[reactant.name]
            self.reactions[reaction.name][reactant] -= int(stoich_value)

        for product, stoich_value in reaction.products.items():
            if isinstance(product, Species):
                product = self.species_names[product.name]
            self.reactions[reaction.name][product] += int(stoich_value)

        return self

    def use_rate_rule(self, rate_rule: "RateRule") -> "SanitizedModel":
        """
        Attach the given rate rule to the sanitized model.
        The rate rule will automatically be validated and sanitized before being applied.

        :param rate_rule: GillesPy2 RateRule object to attach to the sanitized model.
        :type rate_rule: gillespy2.RateRule

        :returns: Pass-through of sanitized model object.
        :rtype: SanitizedModel
        """

        if isinstance(rate_rule.variable, Species):
            variable = rate_rule.variable
        elif not isinstance(rate_rule.variable, Parameter) and \
                    rate_rule.variable in self.model.listOfSpecies.keys():
            variable = self.model.get_species(rate_rule.variable)
        else:
            errmsg = """
            Parameters are not valid variables for the TauHybridCSolver.

            In order to use this variable it will need to be a gillespy2.Species.
            """
            raise SimulationError(errmsg)

        if variable.name in self.species_names:
            sanitized_name = self.species_names.get(variable.name)
            if sanitized_name in self.rate_rules:
                log.warning(f"Duplicate rate rule variable found in C++ solver: {variable}")
            rr_sanitized = self.expr.getexpr_cpp(rate_rule.formula)
            if rr_sanitized is not None:
                self.rate_rules[sanitized_name] = rr_sanitized
            else:
                log.warning(f"Could not sanitize rate rule formula expression: {rate_rule.formula}")
        return self

    def get_template(self, variable=False) -> "dict[str, str]":
        """
        Creates a dictionary of C++ macro definitions from the given model.
        The keys of the dictionary contain the name of the macro definition.
        The values of the dictionary are the values their corresponding macro should be defined to.

        :param variable: Set to true to allow for non-constant parameter values.
        :type variable: bool

        :returns: Dictionary of fully-formatted macro definitions.
        :rtype: dict[str, str]
        """
        results = dict({})

        # Get definitions for variables
        parameter_definitions = template_def_variables(self, variable)
        results.update(parameter_definitions)

        # Get definitions for species
        species_definitions = template_def_species(self)
        results.update(species_definitions)

        # Get definitions for reactions
        reaction_definitions = template_def_reactions(self)
        results.update(reaction_definitions)

        # Get definitions for propensities
        stoch_propensity_definitions = template_def_propensities(self, ode=False)
        ode_propensity_definitions = template_def_propensities(self, ode=True)
        results.update(stoch_propensity_definitions)
        results.update(ode_propensity_definitions)

        return results

    def get_options(self) -> "Optional[dict[str, str]]":
        """
        Creates a dictionary of C++ macro definitions for optional parameters of the model.
        The keys of the dictionary contain the name of the macro definition.
        The values of the dictionary are the values their corresponding macro should be defined to.

        :returns: Dictionary of fully-formatted macro definitions.
        :rtype: dict[str, str]
        """
        options = update_model_options(self, self.options.copy())
        return options if len(options) > 0 else None


def write_template(path: str, model: Model, variable=False):
    """
    Write template definitions to a specified path. If the file does not exist, one is created.
    Definitions are written in `#define KEY VALUES` format.

    :param model: Model to get the macro definitions of.
    :type model: gillespy2.Model

    :param variable: Set to true to allow for non-constant parameter values.
    :type variable: bool
    """

    # Get a dictionary of model defines and transform into a list of strings in
    # `#define KEY VALUE` format.
    defines = get_model_defines(model, variable)
    write_definitions(path, defines)


def write_definitions(path: str, defines: "dict[str, str]"):
    """
    Write the given key-value pairs as a C/C++ template file.
    Contents of the given `defines` dict are written as `#define {key} {value}` format.

    :param path: Absolute filepath of the header file to create (including the `.h` extension.
    :type path: str

    :param defines: Dictionary of key-value pairs representing macro definitions to create.
    :type defines: dict[str, str]
    """
    # Definition dict is transformed into a list of C++ macro definitions, with:
    # `#define KEY VALUE` format.
    template_lines = [(f"#define {key} {value}\n") for key, value in defines.items()]

    # Write generated lines to the template file.
    with open(path, "w") as template_file:
        template_file.writelines(template_lines)


def get_model_defines(model: Model, variable=False) -> "dict[str, str]":
    """
    Creates a dictionary of C++ macro definitions from the given model.
    The keys of the dictionary contain the name of the macro definition.
    The values of the dictionary are the values their corresponding macro should be defined to.

    :param model: Model to get the macro definitions of.
    :type model: gillespy2.Model

    :param variable: Set to true to allow for non-constant parameter values.
    :type variable: bool

    :returns: Dictionary of fully-formatted macro definitions.
    """
    results = {}
    sanitized_model = SanitizedModel(model)

    # Get definitions for variables
    parameter_definitions = template_def_variables(sanitized_model, variable)
    results.update(parameter_definitions)

    # Get definitions for species
    species_definitions = template_def_species(sanitized_model)
    results.update(species_definitions)

    # Get definitions for reactions
    reaction_definitions = template_def_reactions(sanitized_model)
    results.update(reaction_definitions)

    # Get definitions for propensities
    stoch_propensity_definitions = template_def_propensities(sanitized_model, ode=False)
    ode_propensity_definitions = template_def_propensities(sanitized_model, ode=True)
    results.update(stoch_propensity_definitions)
    results.update(ode_propensity_definitions)

    return results


def update_model_options(model: "SanitizedModel", definitions: "dict[str, str]" = None) -> "dict[str, str]":
    """
    Generate the solver-specific options of the `SanitizedModel` with the given definitions.
    This includes both solver-defined custom definitions, as well as SBML features.

    :param model: Sanitized model to populate the options into.
    The model itself is not modified.
    :type model: SanitizedModel

    :param definitions: Dictionary of macro definition key-value pairs to pass to the model.
    :type definitions: dict[str, str]

    :returns: Updated dictionary of macro definitions.
    Includes both solver-specific definitions provided and SBML features on the model.
    :rtype: dict[str, str]
    """
    if definitions is None:
        definitions = {}

    if len(model.rate_rules) > 0:
        definitions.update(template_def_rate_rules(model))

    return definitions


def template_def_rate_rules(model: SanitizedModel) -> "dict[str, str]":
    """
    Generates template definitions for SBML rate rules.
    """
    rr_set = []
    for spec_i, species in enumerate(model.species_names.values()):
        if species in model.rate_rules:
            spec_rr = model.rate_rules.get(species)
            rr_set.append(f"RATE_RULE({spec_i}, {spec_rr})")

    return {
        "GPY_RATE_RULES": " ".join(rr_set)
    }

def template_def_variables(model: SanitizedModel, variable=False) -> "dict[str, str]":
    """
    Formats the relevant parameters to be passed to a C++ simulation template.
    Passed dictionaries/lists are assumed to be sanitized and sorted.

    :param model: Sanitized model containing runtime definitions.
    :type model: SanitizedModel

    :param variable: Indicates whether the model's variables should be modifiable at runtime.
    :type variable: bool

    :returns: The result as a list of tuples containing key-value pairs to be templated.
    """
    # Entries get defined as constant if variable is set to true.
    parameter_type = "VARIABLE" if variable else "CONSTANT"
    # Parameter entries, parsed and formatted
    parameter_set = []
    for param_name, parameter in model.parameters.items():
        parameter_set.append(f"{parameter_type}({param_name},{parameter.expression})")

    return {
        "GPY_PARAMETER_VALUES": " ".join(parameter_set)
    }


def template_def_species(model: SanitizedModel) -> "dict[str, str]":
    """
    Passed dictionaries/lists are assumed to be sanitized and sorted.
    Formats the relevant species data to be passed to a C++ simulation template.

    :param model: Sanitized model containing runtime definitions.
    :type model: SanitizedModel

    :returns: Dictionary of macro definitions for species and data related to species.
    """
    # Parse and format species initial populations
    num_species = len(model.species)
    populations = OrderedDict()

    for spec_name, spec in model.species.items():
        populations[spec_name] = str(spec.initial_value)
    # Species names, parsed and formatted
    sanitized_names = [f"SPECIES_NAME({name})" for name in populations.keys()]
    populations = f"{{{','.join(populations.values())}}}"

    # Match each parameter with its macro definition name
    return {
        "GPY_INIT_POPULATIONS": populations,
        "GPY_NUM_SPECIES": num_species,
        "GPY_SPECIES_NAMES": " ".join(sanitized_names)
    }


def template_def_reactions(model: SanitizedModel, ode=False) -> "dict[str, str]":
    """
    Passed dictionaries/lists are assumed to be sanitized and sorted.
    Formats the relevant reactions and propensities to be passed to a C++ simulation template.

    :param model: Sanitized model containing runtime definitions.
    :type model: SanitizedModel

    :returns: Dictionary of macro definitions for reactions.
    """
    num_reactions = str(len(model.reactions))
    reaction_set = OrderedDict()

    for rxn_name, reaction in model.reactions.items():
        stoich = [str(int(reaction[species])) for species in model.species_names.values()]
        reaction_set[rxn_name] = f"{{{','.join(stoich)}}}"

    reaction_names = " ".join([f"REACTION_NAME({rxn})" for rxn in reaction_set.keys()])
    reaction_set = f"{{{','.join(reaction_set.values())}}}"

    return {
        "GPY_NUM_REACTIONS": num_reactions,
        "GPY_REACTIONS": reaction_set,
        "GPY_REACTION_NAMES": reaction_names,
    }


def template_def_propensities(model: SanitizedModel, ode=False) -> "dict[str, str]":
    """
    Formats the given list of pre-sorted, pre-sanitized propensities.

    :param model: Sanitized model containing runtime definitions.
    :type model: SanitizedModel

    :returns: Dictionary containing propensity macro definitions.
    """
    def_keyword = "ODE_PROPENSITIES" if ode else "PROPENSITIES"
    propensities = model.ode_propensities if ode else model.propensities
    propensities = [f"PROPENSITY({rxn_i},{propensities[rxn]})" for rxn_i, rxn in enumerate(model.reactions.keys())]

    return {
        f"GPY_{def_keyword}": " ".join(propensities)
    }



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import unittest
import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

from profiling import c_profiler

from example_models import Oregonator 
from example_models import VilarOscillator
from example_models import Tyson2StateOscillator 

from gillespy2.solvers.cpp import SSACSolver
from gillespy2.solvers.cpp import ODECSolver
from gillespy2.solvers.cpp import TauLeapingCSolver


class MyTestCase(unittest.TestCase):
    def test_print_profiler_results(self):
        for solver, models in self.solvers.items():
            print(f"=== === === Solver: {solver.name} === === ===")
            for model in models:
                perf_results = c_profiler.run_profiler(model, solver(model=model))
                print(f"  === === Model: {model.name} === ===")
                print(perf_results)
        print(f"=== === === === === === === === === === === === ===")

    def setUp(self) -> None:
        self.solvers = {
            SSACSolver: [
                Tyson2StateOscillator()
            ],
            ODECSolver: [
                Oregonator()
            ],
            TauLeapingCSolver: [
                VilarOscillator()
            ],
        }


if __name__ == '__main__':
    unittest.main()



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import unittest
import tempfile
import os
import re
from example_models import RobustModel
from gillespy2.core.model import import_SBML, export_SBML
from gillespy2.core import gillespyError
from gillespy2 import ODESolver


class TestSBML(unittest.TestCase):

    converted_symbols = {
        "ln": "log",
    }
    test_formula_name = "ar_t_ave"

    def test_sbml_conversion(self):
        try:
            import libsbml
        except ImportError:
            return

        model_path = os.path.join(os.path.dirname(__file__), "assets", "test_sbml.xml")
        sbml_model, errors = import_SBML(model_path)
        sbml_results = sbml_model.run(solver=ODESolver)

    def test_sbml_string_replace(self):
        """
        Ensure that non-supported symbols are converted to supported symbols (such as ln -> log)
        """
        try:
            import libsbml

            # Model is modified from BIOMD0000000012, with species names modified to force possible collisions.
            # For example, species name `PY` -> `xlnx` which, if not implemented correctly,
            #   would cause a collision with `ln` and result in the name `xlogx`, which is wrong.
            model_path = os.path.join(os.path.dirname(__file__), "assets", "math_test_model.xml")
            sbml_model, errors = import_SBML(model_path)
            test_rule = sbml_model.listOfAssignmentRules.get(TestSBML.test_formula_name).formula

            for before, after in TestSBML.converted_symbols.items():
                with self.subTest("Testing SBML conversion", before=before, after=after):
                    self.assertTrue(f"x{before}x" in sbml_model.listOfSpecies,
                                    "Species name 'xlnx' was unexpectedly modified during SBML import")
                    self.assertIsNone(re.search(rf'\b{before}\b', test_rule),
                                      f"Expected SBML formula '{test_rule}' to remove all instances of '{before}'")
                    self.assertIsNotNone(re.search(rf'\b{after}\b', test_rule),
                                         f"Expected SBML formula '{test_rule} to convert all instances of '{after}'")
        except ImportError:
            return

    def test_sbml_export_conversion(self):
        try:
            import libsbml
        except ImportError:
            return

        tmp = tempfile.NamedTemporaryFile(delete=False)
        model = RobustModel()
        path = export_SBML(model, filename=tmp.name)
        document = libsbml.readSBML(path)
        if document.getNumErrors() > 0:
            error_id = document.getError(0).getErrorId()
            file_errors = [libsbml.XMLFileUnreadable, libsbml.XMLFileOperationError]
            assert error_id not in file_errors

    def test_sbml_model_without_propensities(self):
        try:
            import libsbml
        except ImportError:
            return

        model_path = os.path.join(os.path.dirname(__file__), "assets", "model_without_propensities.xml")

        try:
            import_SBML(model_path)
        except gillespyError.InvalidModelError:
            return
        except gillespyError.ParameterError:
            return

        self.fail()


if __name__ == '__main__':
    unittest.main()
